/*
 *	NVidia injector
 *
 *	Copyright (C) 2009	Jasmin Fazlic, iNDi
 *
 *	NVidia injector modified by Fabio (ErmaC) on May 2012,
 *	for allow the cosmetics injection also based on SubVendorID and SubDeviceID.
 *
 *	NVidia injector is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, either version 3 of the License, or
 *	(at your option) any later version.
 *
 *	NVidia driver and injector is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with NVidia injector.	 If not, see <http://www.gnu.org/licenses/>.
 *
 *	Alternatively you can choose to comply with APSL
 *
 *	DCB-Table parsing is based on software (nouveau driver) originally distributed under following license:
 *
 *
 *	Copyright 2005-2006 Erik Waling
 *	Copyright 2006 Stephane Marchesin
 *	Copyright 2007-2009 Stuart Bennett
 *
 *	Permission is hereby granted, free of charge, to any person obtaining a
 *	copy of this software and associated documentation files (the "Software"),
 *	to deal in the Software without restriction, including without limitation
 *	the rights to use, copy, modify, merge, publish, distribute, sublicense,
 *	and/or sell copies of the Software, and to permit persons to whom the
 *	Software is furnished to do so, subject to the following conditions:
 *
 *	The above copyright notice and this permission notice shall be included in
 *	all copies or substantial portions of the Software.
 *
 *	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 *	THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 *	WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF
 *	OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 *	SOFTWARE.
 */

#include "boot.h"
#include "bootstruct.h"
#include "pci.h"
#include "platform.h"
#include "device_inject.h"
#include "nvidia.h"
#include "nvidia_helper.h"

#ifndef DEBUG_NVIDIA
#define DEBUG_NVIDIA 0
#endif

#if DEBUG_NVIDIA
#define DBG(x...)	printf(x)
#else
#define DBG(x...)
#endif

#define NVIDIA_ROM_SIZE				0x20000
#define PATCH_ROM_SUCCESS			1
#define PATCH_ROM_SUCCESS_HAS_LVDS		2
#define PATCH_ROM_FAILED			0
#define MAX_NUM_DCB_ENTRIES			16
#define TYPE_GROUPED				0xff
#define READ_BYTE(rom, offset) (*(u_char *)(rom + offset))
#define READ_LE_SHORT(rom, offset) (READ_BYTE(rom, offset+1) << 8 | READ_BYTE(rom, offset))
#define READ_LE_INT(rom, offset)   (READ_LE_SHORT(rom, offset+2) << 16 | READ_LE_SHORT(rom, offset))
#define WRITE_LE_SHORT(data)       (((data) << 8 & 0xff00) | ((data) >> 8 & 0x00ff ))
#define WRITE_LE_INT(data)         (WRITE_LE_SHORT(data) << 16 | WRITE_LE_SHORT(data >> 16))

//static bool     showGeneric         = false;
char generic_name[128];
extern uint32_t devices_number;

const char *nvidia_compatible_0[]       =	{ "@0,compatible",	"NVDA,NVMac"	 };
const char *nvidia_compatible_1[]       =	{ "@1,compatible",	"NVDA,NVMac"	 };
const char *nvidia_device_type_0[]      =	{ "@0,device_type",	"display"	 };
const char *nvidia_device_type_1[]      =	{ "@1,device_type",	"display"	 };
const char *nvidia_device_type_parent[] =	{ "device_type",	"NVDA,Parent"	 };
const char *nvidia_device_type_child[]	=	{ "device_type",	"NVDA,Child"	 };
const char *nvidia_name_0[]             =	{ "@0,name",		"NVDA,Display-A" };
const char *nvidia_name_1[]             =	{ "@1,name",		"NVDA,Display-B" };
const char *nvidia_slot_name[]          =	{ "AAPL,slot-name", "Slot-1"		 };

static uint8_t default_NVCAP[]= {
	0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0d, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0a,
	0x00, 0x00, 0x00, 0x00
};

#define NVCAP_LEN ( sizeof(default_NVCAP) / sizeof(uint8_t) )

static uint8_t default_dcfg_0[]		=	{0x03, 0x01, 0x03, 0x00};
static uint8_t default_dcfg_1[]		=	{0xff, 0xff, 0x00, 0x01};

#define DCFG0_LEN ( sizeof(default_dcfg_0) / sizeof(uint8_t) )
#define DCFG1_LEN ( sizeof(default_dcfg_1) / sizeof(uint8_t) )

static uint8_t default_NVPM[]= {
	0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00
};

#define NVPM_LEN ( sizeof(default_NVPM) / sizeof(uint8_t) )

static nvidia_pci_info_t nvidia_card_vendors[] = {
    { 0x10190000,	"Elitegroup" },
	{ 0x10250000,	"Acer" },
	{ 0x10280000,	"Dell" },
	{ 0x10330000,	"NEC" },
	{ 0x103C0000,	"HP" },
	{ 0x10430000,	"Asus" },
	{ 0x104D0000,	"Sony" },
	{ 0x105B0000,	"Foxconn" },
	{ 0x106B0000,	"Apple" },
	{ 0x10710000,	"Mitac" },
	{ 0x107B0000,	"Gateway" },
	{ 0x107D0000,	"Leadtek" },
	{ 0x109F0000,	"Trigem" },
	{ 0x10B00000,	"Gainward" },
	{ 0x10CF0000,	"Fujitsu" },
	{ 0x10DE0000,	"nVidia" },
	{ 0x11790000,	"Toshiba" },
	{ 0x12970000,	"Shuttle" },
	{ 0x13DC0000,	"Netbost" },
	{ 0x144D0000,	"Samsung" },
	{ 0x14580000,	"Gigabyte" },
	{ 0x14620000,	"MSi" },
	{ 0x14C00000,	"Compal" },
	{ 0x152D0000,	"Quanta" },
	{ 0x15540000,	"Prolink" },
	{ 0x15580000,	"Clevo" },
	{ 0x15690000,	"Palit" },
	{ 0x161F0000,	"Arima" },
 	{ 0x16310000,	"NEC" },
	{ 0x16420000,	"Bitland" },
	{ 0x16820000,	"XFX" },
	{ 0x17340000,	"Fujitsu" },
	{ 0x174B0000,	"PC Partner" },
	{ 0x17AA0000,	"Lenovo" },
 	{ 0x17C00000,	"Wistron" },
	{ 0x17FF0000,	"Benq" },
	{ 0x18490000,	"ASRock" },
	{ 0x18540000,	"LG" },
	{ 0x18640000,	"LG" },
	{ 0x18940000,	"LG" },
	{ 0x19610000,	"ESS" },
	{ 0x196E0000,	"PNY" },
	{ 0x19910000,	"Topstar" },
	{ 0x19DA0000,	"Zotac" },
	{ 0x19F10000,	"BFG" },
	{ 0x1ACC0000,	"Point of View" },
	{ 0x1B0A0000,	"Pegatron" },
	{ 0x1B130000,	"Jaton" },
	{ 0x34420000,	"Bihl" },
	{ 0x38420000,	"EVGA" },
	{ 0x73770000,	"Colorful" },
};

static nvidia_pci_info_t nvidia_card_generic[] = {
	// 0000 - 0040	
	{ 0x10DE0000,	"Unknown" },
	// 0040 - 004F	
	{ 0x10DE0040,	"NVIDIA GeForce 6800 Ultra" },
	{ 0x10DE0041,	"NVIDIA GeForce 6800" },
	{ 0x10DE0042,	"NVIDIA GeForce 6800 LE" },
	{ 0x10DE0043,	"NVIDIA GeForce 6800 XE" },
	{ 0x10DE0044,	"NVIDIA GeForce 6800 XT" },
	{ 0x10DE0045,	"NVIDIA GeForce 6800 GT" },
	{ 0x10DE0046,	"NVIDIA GeForce 6800 GT" },
	{ 0x10DE0047,	"NVIDIA GeForce 6800 GS" },
	{ 0x10DE0048,	"NVIDIA GeForce 6800 XT" },
	{ 0x10DE004D,	"NVIDIA Quadro FX 3400" },
	{ 0x10DE004E,	"NVIDIA Quadro FX 4000" },
	// 0050 - 005F
	// 0060 - 006F
	// 0070 - 007F
	// 0080 - 008F
	// 0090 - 009F
	{ 0x10DE0090,	"NVIDIA GeForce 7800 GTX" },
	{ 0x10DE0091,	"NVIDIA GeForce 7800 GTX" },
	{ 0x10DE0092,	"NVIDIA GeForce 7800 GT" },
	{ 0x10DE0093,	"NVIDIA GeForce 7800 GS" },
	{ 0x10DE0095,	"NVIDIA GeForce 7800 SLI" },
	{ 0x10DE0098,	"NVIDIA GeForce Go 7800" },
	{ 0x10DE0099,	"NVIDIA GeForce Go 7800 GTX" },
	{ 0x10DE009D,	"NVIDIA Quadro FX 4500" },
	// 00A0 - 00AF	
	// 00B0 - 00BF	
	// 00C0 - 00CF	
	{ 0x10DE00C0,	"NVIDIA GeForce 6800 GS" },
	{ 0x10DE00C1,	"NVIDIA GeForce 6800" },
	{ 0x10DE00C2,	"NVIDIA GeForce 6800 LE" },
	{ 0x10DE00C3,	"NVIDIA GeForce 6800 XT" },
	{ 0x10DE00C8,	"NVIDIA GeForce Go 6800" },
	{ 0x10DE00C9,	"NVIDIA GeForce Go 6800 Ultra" },
	{ 0x10DE00CC,	"NVIDIA Quadro FX Go1400" },
	{ 0x10DE00CD,	"NVIDIA Quadro FX 3450/4000 SDI" },
	{ 0x10DE00CE,	"NVIDIA Quadro FX 1400" },
	// 00D0 - 00DF
	// 00E0 - 00EF
	// 00F0 - 00FF
	{ 0x10DE00F1,	"NVIDIA GeForce 6600 GT" },
	{ 0x10DE00F2,	"NVIDIA GeForce 6600" },
	{ 0x10DE00F3,	"NVIDIA GeForce 6200" },
	{ 0x10DE00F4,	"NVIDIA GeForce 6600 LE" },
	{ 0x10DE00F5,	"NVIDIA GeForce 7800 GS" },
	{ 0x10DE00F6,	"NVIDIA GeForce 6800 GS/XT" },
	{ 0x10DE00F8,	"NVIDIA Quadro FX 3400/4400" },
	{ 0x10DE00F9,	"NVIDIA GeForce 6800 Series GPU" },
	// 0100 - 010F
	// 0110 - 011F
	// 0120 - 012F
	// 0130 - 013F
	// 0140 - 014F
	{ 0x10DE0140,	"NVIDIA GeForce 6600 GT" },
	{ 0x10DE0141,	"NVIDIA GeForce 6600" },
	{ 0x10DE0142,	"NVIDIA GeForce 6600 LE" },
	{ 0x10DE0143,	"NVIDIA GeForce 6600 VE" },
	{ 0x10DE0144,	"NVIDIA GeForce Go 6600" },
	{ 0x10DE0145,	"NVIDIA GeForce 6610 XL" },
	{ 0x10DE0146,	"NVIDIA GeForce Go 6600 TE/6200 TE" },
	{ 0x10DE0147,	"NVIDIA GeForce 6700 XL" },
	{ 0x10DE0148,	"NVIDIA GeForce Go 6600" },
	{ 0x10DE0149,	"NVIDIA GeForce Go 6600 GT" },
	{ 0x10DE014A,	"NVIDIA Quadro NVS 440" },
	{ 0x10DE014C,	"NVIDIA Quadro FX 550" },
	{ 0x10DE014D,	"NVIDIA Quadro FX 550" },
	{ 0x10DE014E,	"NVIDIA Quadro FX 540" },
	{ 0x10DE014F,	"NVIDIA GeForce 6200" },
	// 0150 - 015F
	// 0160 - 016F
	{ 0x10DE0160,	"NVIDIA GeForce 6500" },
	{ 0x10DE0161,	"NVIDIA GeForce 6200 TurboCache(TM)" },
	{ 0x10DE0162,	"NVIDIA GeForce 6200SE TurboCache(TM)" },
	{ 0x10DE0163,	"NVIDIA GeForce 6200 LE" },
	{ 0x10DE0164,	"NVIDIA GeForce Go 6200" },
	{ 0x10DE0165,	"NVIDIA Quadro NVS 285" },
	{ 0x10DE0166,	"NVIDIA GeForce Go 6400" },
	{ 0x10DE0167,	"NVIDIA GeForce Go 6200" },
	{ 0x10DE0168,	"NVIDIA GeForce Go 6400" },
	{ 0x10DE0169,	"NVIDIA GeForce 6250" },
	{ 0x10DE016A,	"NVIDIA GeForce 7100 GS" },
	{ 0x10DE016C,	"NVIDIA NV44GLM" },
	{ 0x10DE016D,	"NVIDIA NV44GLM" },
	// 0170 - 017F
	// 0180 - 018F
	// 0190 - 019F
	{ 0x10DE0191,	"NVIDIA GeForce 8800 GTX" },
	{ 0x10DE0193,	"NVIDIA GeForce 8800 GTS" },
	{ 0x10DE0194,	"NVIDIA GeForce 8800 Ultra" },
	{ 0x10DE0197,	"NVIDIA Tesla C870" },
	{ 0x10DE019D,	"NVIDIA Quadro FX 5600" },
	{ 0x10DE019E,	"NVIDIA Quadro FX 4600" },
	// 01A0 - 01AF
	// 01B0 - 01BF
	// 01C0 - 01CF
	// 01D0 - 01DF
	{ 0x10DE01D0,	"NVIDIA GeForce 7350 LE" },
	{ 0x10DE01D1,	"NVIDIA GeForce 7300 LE" },
	{ 0x10DE01D2,	"NVIDIA GeForce 7550 LE" },
	{ 0x10DE01D3,	"NVIDIA GeForce 7300 SE/7200 GS" },
	{ 0x10DE01D6,	"NVIDIA GeForce Go 7200" },
	{ 0x10DE01D7,	"NVIDIA GeForce Go 7300" },
	{ 0x10DE01D8,	"NVIDIA GeForce Go 7400" },
	{ 0x10DE01D9,	"NVIDIA GeForce Go 7450" },
	{ 0x10DE01DA,	"NVIDIA Quadro NVS 110M" },
	{ 0x10DE01DB,	"NVIDIA Quadro NVS 120M" },
	{ 0x10DE01DC,	"NVIDIA Quadro FX 350M" },
	{ 0x10DE01DD,	"NVIDIA GeForce 7500 LE" },
	{ 0x10DE01DE,	"NVIDIA Quadro FX 350" },
	{ 0x10DE01DF,	"NVIDIA GeForce 7300 GS" },
	// 01E0 - 01EF
	// 01F0 - 01FF
	{ 0x10DE01F0,	"NVIDIA GeForce4 MX" },
	// 0200 - 020F
	// 0210 - 021F
	{ 0x10DE0211,	"NVIDIA GeForce 6800" },
	{ 0x10DE0212,	"NVIDIA GeForce 6800 LE" },
	{ 0x10DE0215,	"NVIDIA GeForce 6800 GT" },
	{ 0x10DE0218,	"NVIDIA GeForce 6800 XT" },
	// 0220 - 022F
	{ 0x10DE0221,	"NVIDIA GeForce 6200" },
	{ 0x10DE0222,	"NVIDIA GeForce 6200 A-LE" },
	{ 0x10DE0228,	"NVIDIA NV44M" },
	// 0230 - 023F
	// 0240 - 024F
	{ 0x10DE0240,	"NVIDIA GeForce 6150" },
	{ 0x10DE0241,	"NVIDIA GeForce 6150 LE" },
	{ 0x10DE0242,	"NVIDIA GeForce 6100" },
	{ 0x10DE0243,	"NVIDIA C51" },
	{ 0x10DE0244,	"NVIDIA GeForce Go 6150" },
	{ 0x10DE0245,	"NVIDIA Quadro NVS 210S / GeForce 6150LE" },
	{ 0x10DE0247,	"NVIDIA GeForce Go 6100" },
	// 0250 - 025F
	{ 0x10DE025B,	"NVIDIA Quadro4 700 XGL" },
	// 0260 - 026F
	// 0270 - 027F
	// 0280 - 028F
	// 0290 - 029F
	{ 0x10DE0290,	"NVIDIA GeForce 7900 GTX" },
	{ 0x10DE0291,	"NVIDIA GeForce 7900 GT/GTO" },
	{ 0x10DE0292,	"NVIDIA GeForce 7900 GS" },
	{ 0x10DE0293,	"NVIDIA GeForce 7950 GX2" },
	{ 0x10DE0294,	"NVIDIA GeForce 7950 GX2" },
	{ 0x10DE0295,	"NVIDIA GeForce 7950 GT" },
	{ 0x10DE0298,	"NVIDIA GeForce Go 7900 GS" },
	{ 0x10DE0299,	"NVIDIA GeForce Go 7900 GTX" },
	{ 0x10DE029A,	"NVIDIA Quadro FX 2500M" },
	{ 0x10DE029B,	"NVIDIA Quadro FX 1500M" },
	{ 0x10DE029C,	"NVIDIA Quadro FX 5500" },
	{ 0x10DE029D,	"NVIDIA Quadro FX 3500" },
	{ 0x10DE029E,	"NVIDIA Quadro FX 1500" },
	{ 0x10DE029F,	"NVIDIA Quadro FX 4500 X2" },
	// 02A0 - 02AF
	// 02B0 - 02BF
	// 02C0 - 02CF
	// 02D0 - 02DF
	// 02E0 - 02EF
	{ 0x10DE02E0,	"NVIDIA GeForce 7600 GT" },
	{ 0x10DE02E1,	"NVIDIA GeForce 7600 GS" },
	{ 0x10DE02E2,	"NVIDIA GeForce 7300 GT" },
	{ 0x10DE02E3,	"NVIDIA GeForce 7900 GS" },
	{ 0x10DE02E4,	"NVIDIA GeForce 7950 GT" },
	// 02F0 - 02FF
	// 0300 - 030F
	{ 0x10DE0301,	"NVIDIA GeForce FX 5800 Ultra" },
	{ 0x10DE0302,	"NVIDIA GeForce FX 5800" },
	{ 0x10DE0308,	"NVIDIA Quadro FX 2000" },
	{ 0x10DE0309,	"NVIDIA Quadro FX 1000" },
	// 0310 - 031F
	{ 0x10DE0311,	"NVIDIA GeForce FX 5600 Ultra" },
	{ 0x10DE0312,	"NVIDIA GeForce FX 5600" },
	{ 0x10DE0314,	"NVIDIA GeForce FX 5600XT" },
	{ 0x10DE031A,	"NVIDIA GeForce FX Go5600" },
	{ 0x10DE031B,	"NVIDIA GeForce FX Go5650" },
	{ 0x10DE031C,	"NVIDIA Quadro FX Go700" },
	// 0320 - 032F
	{ 0x10DE0320,	"NVIDIA GeForce FX 5200" },
	{ 0x10DE0321,	"NVIDIA GeForce FX 5200 Ultra" },
	{ 0x10DE0322,	"NVIDIA GeForce FX 5200" },
	{ 0x10DE0323,	"NVIDIA GeForce FX 5200 LE" },
	{ 0x10DE0324,	"NVIDIA GeForce FX Go5200" },
	{ 0x10DE0325,	"NVIDIA GeForce FX Go5250" },
	{ 0x10DE0326,	"NVIDIA GeForce FX 5500" },
	{ 0x10DE0328,	"NVIDIA GeForce FX Go5200 32M/64M" },
	{ 0x10DE0329,	"NVIDIA GeForce FX Go5200" },
	{ 0x10DE032A,	"NVIDIA Quadro NVS 55/280 PCI" },
	{ 0x10DE032B,	"NVIDIA Quadro FX 500/600 PCI" },
	{ 0x10DE032C,	"NVIDIA GeForce FX Go53xx Series" },
	{ 0x10DE032D,	"NVIDIA GeForce FX Go5100" },
	{ 0x10DE032F,	"NV34GL" },
	// 0330 - 033F
	{ 0x10DE0330,	"NVIDIA GeForce FX 5900 Ultra" },
	{ 0x10DE0331,	"NVIDIA GeForce FX 5900" },
	{ 0x10DE0332,	"NVIDIA GeForce FX 5900XT" },
	{ 0x10DE0333,	"NVIDIA GeForce FX 5950 Ultra" },
	{ 0x10DE0334,	"NVIDIA GeForce FX 5900ZT" },
	{ 0x10DE0338,	"NVIDIA Quadro FX 3000" },
	{ 0x10DE033F,	"NVIDIA Quadro FX 700" },
	// 0340 - 034F
	{ 0x10DE0341,	"NVIDIA GeForce FX 5700 Ultra" },
	{ 0x10DE0342,	"NVIDIA GeForce FX 5700" },
	{ 0x10DE0343,	"NVIDIA GeForce FX 5700LE" },
	{ 0x10DE0344,	"NVIDIA GeForce FX 5700VE" },
	{ 0x10DE0345,	"NV36.5" },
	{ 0x10DE0347,	"NVIDIA GeForce FX Go5700" },
	{ 0x10DE0348,	"NVIDIA GeForce FX Go5700" },
	{ 0x10DE0349,	"NV36M Pro" },
	{ 0x10DE034B,	"NV36MAP" },
	{ 0x10DE034C,	"NVIDIA Quadro FX Go1000" },
	{ 0x10DE034E,	"NVIDIA Quadro FX 1100" },
	{ 0x10DE034F,	"NV36GL" },
	// 0350 - 035F
	// 0360 - 036F
	// 0370 - 037F
	// 0380 - 038F
	{ 0x10DE038B,	"NVIDIA GeForce 7650 GS" },
	// 0390 - 039F
	{ 0x10DE0390,	"NVIDIA GeForce 7650 GS" },
	{ 0x10DE0391,	"NVIDIA GeForce 7600 GT" },
	{ 0x10DE0392,	"NVIDIA GeForce 7600 GS" },
	{ 0x10DE0393,	"NVIDIA GeForce 7300 GT" },
	{ 0x10DE0394,	"NVIDIA GeForce 7600 LE" },
	{ 0x10DE0395,	"NVIDIA GeForce 7300 GT" },
	{ 0x10DE0397,	"NVIDIA GeForce Go 7700" },
	{ 0x10DE0398,	"NVIDIA GeForce Go 7600" },
	{ 0x10DE0399,	"NVIDIA GeForce Go 7600 GT"},
	{ 0x10DE039A,	"NVIDIA Quadro NVS 300M" },
	{ 0x10DE039B,	"NVIDIA GeForce Go 7900 SE" },
	{ 0x10DE039C,	"NVIDIA Quadro FX 560M" },
	{ 0x10DE039E,	"NVIDIA Quadro FX 560" },
	// 03A0 - 03AF
	// 03B0 - 03BF
	// 03C0 - 03CF
	// 03D0 - 03DF
	{ 0x10DE03D0,	"NVIDIA GeForce 6150SE nForce 430" },
	{ 0x10DE03D1,	"NVIDIA GeForce 6100 nForce 405" },
	{ 0x10DE03D2,	"NVIDIA GeForce 6100 nForce 400" },
	{ 0x10DE03D5,	"NVIDIA GeForce 6100 nForce 420" },
	{ 0x10DE03D6,	"NVIDIA GeForce 7025 / nForce 630a" },
	// 03E0 - 03EF
	// 03F0 - 03FF
	// 0400 - 040F
	{ 0x10DE0400,	"NVIDIA GeForce 8600 GTS" },
	{ 0x10DE0401,	"NVIDIA GeForce 8600 GT" },
	{ 0x10DE0402,	"NVIDIA GeForce 8600 GT" },
	{ 0x10DE0403,	"NVIDIA GeForce 8600 GS" },
	{ 0x10DE0404,	"NVIDIA GeForce 8400 GS" },
	{ 0x10DE0405,	"NVIDIA GeForce 9500M GS" },
	{ 0x10DE0406,	"NVIDIA GeForce 8300 GS" },
	{ 0x10DE0407,	"NVIDIA GeForce 8600M GT" },
	{ 0x10DE0408,	"NVIDIA GeForce 9650M GS" },
	{ 0x10DE0409,	"NVIDIA GeForce 8700M GT" },
	{ 0x10DE040A,	"NVIDIA Quadro FX 370" },
	{ 0x10DE040B,	"NVIDIA Quadro NVS 320M" },
	{ 0x10DE040C,	"NVIDIA Quadro FX 570M" },
	{ 0x10DE040D,	"NVIDIA Quadro FX 1600M" },
	{ 0x10DE040E,	"NVIDIA Quadro FX 570" },
	{ 0x10DE040F,	"NVIDIA Quadro FX 1700" },
	// 0410 - 041F
	{ 0x10DE0410,	"NVIDIA GeForce GT 330" },
	// 0420 - 042F
	{ 0x10DE0420,	"NVIDIA GeForce 8400 SE" },
	{ 0x10DE0421,	"NVIDIA GeForce 8500 GT" },
	{ 0x10DE0422,	"NVIDIA GeForce 8400 GS" },
	{ 0x10DE0423,	"NVIDIA GeForce 8300 GS" },
	{ 0x10DE0424,	"NVIDIA GeForce 8400 GS" },
	{ 0x10DE0425,	"NVIDIA GeForce 8600M GS" },
	{ 0x10DE0426,	"NVIDIA GeForce 8400M GT" },
	{ 0x10DE0427,	"NVIDIA GeForce 8400M GS" },
	{ 0x10DE0428,	"NVIDIA GeForce 8400M G" },
	{ 0x10DE0429,	"NVIDIA Quadro NVS 140M" },
	{ 0x10DE042A,	"NVIDIA Quadro NVS 130M" },
	{ 0x10DE042B,	"NVIDIA Quadro NVS 135M" },
	{ 0x10DE042C,	"NVIDIA GeForce 9400 GT" },
	{ 0x10DE042D,	"NVIDIA Quadro FX 360M" },
	{ 0x10DE042E,	"NVIDIA GeForce 9300M G" },
	{ 0x10DE042F,	"NVIDIA Quadro NVS 290" },
	// 0430 - 043F
	// 0440 - 044F
	// 0450 - 045F
	// 0460 - 046F
	// 0470 - 047F
	// 0480 - 048F
	// 0490 - 049F
	// 04A0 - 04AF
	// 04B0 - 04BF
	// 04C0 - 04CF
	{ 0x10DE04C0,	"NVIDIA G78" },
	{ 0x10DE04C1,	"NVIDIA G78" },
	{ 0x10DE04C2,	"NVIDIA G78" },
	{ 0x10DE04C3,	"NVIDIA G78" },
	{ 0x10DE04C4,	"NVIDIA G78" },
	{ 0x10DE04C5,	"NVIDIA G78" },
	{ 0x10DE04C6,	"NVIDIA G78" },
	{ 0x10DE04C7,	"NVIDIA G78" },
	{ 0x10DE04C8,	"NVIDIA G78" },
	{ 0x10DE04C9,	"NVIDIA G78" },
	{ 0x10DE04CA,	"NVIDIA G78" },
	{ 0x10DE04CB,	"NVIDIA G78" },
	{ 0x10DE04CC,	"NVIDIA G78" },
	{ 0x10DE04CD,	"NVIDIA G78" },
	{ 0x10DE04CE,	"NVIDIA G78" },
	{ 0x10DE04CF,	"NVIDIA G78" },
	// 04D0 - 04DF
	// 04E0 - 04EF
	// 04F0 - 04FF
	// 0500 - 050F
	// 0510 - 051F
	// 0520 - 052F
	// 0530 - 053F
	{ 0x10DE0530,	"NVIDIA GeForce 7190M / nForce 650M" },
	{ 0x10DE0531,	"NVIDIA GeForce 7150M / nForce 630M" },
	{ 0x10DE0533,	"NVIDIA GeForce 7000M / nForce 610M" },
	{ 0x10DE053A,	"NVIDIA GeForce 7050 PV / nForce 630a" },
	{ 0x10DE053B,	"NVIDIA GeForce 7050 PV / nForce 630a" },
	{ 0x10DE053E,	"NVIDIA GeForce 7025 / nForce 630a" },
	// 0540 - 054F
	// 0550 - 055F
	// 0560 - 056F
	// 0570 - 057F
	// 0580 - 058F
	// 0590 - 059F
	// 05A0 - 05AF
	// 05B0 - 05BF
	// 05C0 - 05CF
	// 05D0 - 05DF
	// 05E0 - 05EF
	{ 0x10DE05E0,	"NVIDIA GeForce GTX 295" },
	{ 0x10DE05E1,	"NVIDIA GeForce GTX 280" },
	{ 0x10DE05E2,	"NVIDIA GeForce GTX 260" },
	{ 0x10DE05E3,	"NVIDIA GeForce GTX 285" },
	{ 0x10DE05E4,	"NVIDIA GT200" },
	{ 0x10DE05E5,	"NVIDIA GT200" },
	{ 0x10DE05E6,	"NVIDIA GeForce GTX 275" },
	{ 0x10DE05E7,	"NVIDIA Tesla C1060" },
	{ 0x10DE05E8,	"NVIDIA GT200" },
	{ 0x10DE05E9,	"NVIDIA GT200" },
	{ 0x10DE05EA,	"NVIDIA GeForce GTX 260" },
	{ 0x10DE05EB,	"NVIDIA GeForce GTX 295" },
	{ 0x10DE05EC,	"NVIDIA GT200" },
	{ 0x10DE05ED,	"NVIDIA Quadroplex 2200 D2" },
	{ 0x10DE05EE,	"NVIDIA GT200" },
	{ 0x10DE05EF,	"NVIDIA GT200" },
	// 05F0 - 05FF
	{ 0x10DE05F0,	"NVIDIA GT200" },
	{ 0x10DE05F1,	"NVIDIA GT200" },
	{ 0x10DE05F2,	"NVIDIA GT200" },
	{ 0x10DE05F3,	"NVIDIA GT200" },
	{ 0x10DE05F4,	"NVIDIA GT200" },
	{ 0x10DE05F5,	"NVIDIA GT200" },
	{ 0x10DE05F6,	"NVIDIA GT200" },
	{ 0x10DE05F7,	"NVIDIA GT200" },
	{ 0x10DE05F8,	"NVIDIA Quadroplex 2200 S4" },
	{ 0x10DE05F9,	"NVIDIA Quadro CX" },
	{ 0x10DE05FA,	"NVIDIA GT200" },
	{ 0x10DE05FB,	"NVIDIA GT200" },
	{ 0x10DE05FC,	"NVIDIA GT200" },
	{ 0x10DE05FD,	"NVIDIA Quadro FX 5800" },
	{ 0x10DE05FE,	"NVIDIA Quadro FX 4800" },
	{ 0x10DE05FF,	"NVIDIA Quadro FX 3800" },
	// 0600 - 060F
	{ 0x10DE0600,	"NVIDIA GeForce 8800 GTS" },
	{ 0x10DE0601,	"NVIDIA GeForce 9800 GT" },
	{ 0x10DE0602,	"NVIDIA GeForce 8800 GT" },
	{ 0x10DE0603,	"NVIDIA GeForce GT 230" },
	{ 0x10DE0604,	"NVIDIA GeForce 9800 GX2" },
	{ 0x10DE0605,	"NVIDIA GeForce 9800 GT" },
	{ 0x10DE0606,	"NVIDIA GeForce 8800 GS" },
	{ 0x10DE0607,	"NVIDIA GeForce GTS 240" },
	{ 0x10DE0608,	"NVIDIA GeForce 9800M GTX" },
	{ 0x10DE0609,	"NVIDIA GeForce 8800M GTS" },
	{ 0x10DE060A,	"NVIDIA GeForce GTX 280M" },
	{ 0x10DE060B,	"NVIDIA GeForce 9800M GT" },
	{ 0x10DE060C,	"NVIDIA GeForce 8800M GTX" },
	{ 0x10DE060D,	"NVIDIA GeForce 8800 GS" },
	{ 0x10DE060F,	"NVIDIA GeForce GTX 285M" },
	// 0610 - 061F
	{ 0x10DE0610,	"NVIDIA GeForce 9600 GSO" },
	{ 0x10DE0611,	"NVIDIA GeForce 8800 GT" },
	{ 0x10DE0612,	"NVIDIA GeForce 9800 GTX" },
	{ 0x10DE0613,	"NVIDIA GeForce 9800 GTX+" },
	{ 0x10DE0614,	"NVIDIA GeForce 9800 GT" },
	{ 0x10DE0615,	"NVIDIA GeForce GTS 250" },
	{ 0x10DE0617,	"NVIDIA GeForce 9800M GTX" },
	{ 0x10DE0618,	"NVIDIA GeForce GTX 260M" },
	{ 0x10DE0619,	"NVIDIA Quadro FX 4700 X2" },
	{ 0x10DE061A,	"NVIDIA Quadro FX 3700" },
	{ 0x10DE061B,	"NVIDIA Quadro VX 200" },
	{ 0x10DE061C,	"NVIDIA Quadro FX 3600M" },
	{ 0x10DE061D,	"NVIDIA Quadro FX 2800M" },
	{ 0x10DE061E,	"NVIDIA Quadro FX 3700M" },
	{ 0x10DE061F,	"NVIDIA Quadro FX 3800M" },
	// 0620 - 062F
	{ 0x10DE0620,	"NVIDIA G94" },
	{ 0x10DE0621,	"NVIDIA GeForce GT 230" },
	{ 0x10DE0622,	"NVIDIA GeForce 9600 GT" },
	{ 0x10DE0623,	"NVIDIA GeForce 9600 GS" },
	{ 0x10DE0624,	"NVIDIA G94" },
	{ 0x10DE0625,	"NVIDIA GeForce 9600 GSO 512"},
	{ 0x10DE0626,	"NVIDIA GeForce GT 130" },
	{ 0x10DE0627,	"NVIDIA GeForce GT 140" },
	{ 0x10DE0628,	"NVIDIA GeForce 9800M GTS" },
	{ 0x10DE0629,	"NVIDIA G94" },
	{ 0x10DE062A,	"NVIDIA GeForce 9700M GTS" },
	{ 0x10DE062B,	"NVIDIA GeForce 9800M GS" },
	{ 0x10DE062C,	"NVIDIA GeForce 9800M GTS" },
	{ 0x10DE062D,	"NVIDIA GeForce 9600 GT" },
	{ 0x10DE062E,	"NVIDIA GeForce 9600 GT" },
	{ 0x10DE062F,	"NVIDIA GeForce 9800 S" },
	// 0630 - 063F
	{ 0x10DE0630,	"NVIDIA GeForce 9700 S" },
	{ 0x10DE0631,	"NVIDIA GeForce GTS 160M" },
	{ 0x10DE0632,	"NVIDIA GeForce GTS 150M" },
	{ 0x10DE0633,	"NVIDIA G94" },
	{ 0x10DE0634,	"NVIDIA G94" },
	{ 0x10DE0635,	"NVIDIA GeForce 9600 GSO" },
	{ 0x10DE0636,	"NVIDIA G94" },
	{ 0x10DE0637,	"NVIDIA GeForce 9600 GT" },
	{ 0x10DE0638,	"NVIDIA Quadro FX 1800" },
	{ 0x10DE0639,	"NVIDIA G94" },
	{ 0x10DE063A,	"NVIDIA Quadro FX 2700M" },
	{ 0x10DE063B,	"NVIDIA G94" },
	{ 0x10DE063C,	"NVIDIA G94" },
	{ 0x10DE063D,	"NVIDIA G94" },
	{ 0x10DE063E,	"NVIDIA G94" },
	{ 0x10DE063F,	"NVIDIA G94" },
	// 0640 - 064F
	{ 0x10DE0640,	"NVIDIA GeForce 9500 GT" },
	{ 0x10DE0641,	"NVIDIA GeForce 9400 GT" },
	{ 0x10DE0642,	"NVIDIA GeForce 8400 GS" },
	{ 0x10DE0643,	"NVIDIA GeForce 9500 GT" },
	{ 0x10DE0644,	"NVIDIA GeForce 9500 GS" },
	{ 0x10DE0645,	"NVIDIA GeForce 9500 GS" },
	{ 0x10DE0646,	"NVIDIA GeForce GT 120" },
	{ 0x10DE0647,	"NVIDIA GeForce 9600M GT" },
	{ 0x10DE0648,	"NVIDIA GeForce 9600M GS" },
	{ 0x10DE0649,	"NVIDIA GeForce 9600M GT" },
	{ 0x10DE064A,	"NVIDIA GeForce 9700M GT" },
	{ 0x10DE064B,	"NVIDIA GeForce 9500M G" },
	{ 0x10DE064C,	"NVIDIA GeForce 9650M GT" },
	// 0650 - 065F
	{ 0x10DE0650,	"NVIDIA G96-825" },
	{ 0x10DE0651,	"NVIDIA GeForce G 110M" },
	{ 0x10DE0652,	"NVIDIA GeForce GT 130M" },
	{ 0x10DE0653,	"NVIDIA GeForce GT 120M" },
	{ 0x10DE0654,	"NVIDIA GeForce GT 220M" },
	{ 0x10DE0655,	"NVIDIA GeForce GT 120" },
	{ 0x10DE0656,	"NVIDIA GeForce 9650 S" },
	{ 0x10DE0657,	"NVIDIA G96" },
	{ 0x10DE0658,	"NVIDIA Quadro FX 380" },
	{ 0x10DE0659,	"NVIDIA Quadro FX 580" },
	{ 0x10DE065A,	"NVIDIA Quadro FX 1700M" },
	{ 0x10DE065B,	"NVIDIA GeForce 9400 GT" },
	{ 0x10DE065C,	"NVIDIA Quadro FX 770M" },
	{ 0x10DE065D,	"NVIDIA G96" },
	{ 0x10DE065E,	"NVIDIA G96" },
	{ 0x10DE065F,	"NVIDIA GeForce G210" },
	// 0660 - 066F
	// 0670 - 067F
	// 0680 - 068F
	// 0690 - 069F
	// 06A0 - 06AF
	{ 0x10DE06A0,	"NVIDIA GT214" },
	// 06B0 - 06BF
	{ 0x10DE06B0,	"NVIDIA GT214" },
	// 06C0 - 06CF
	{ 0x10DE06C0,	"NVIDIA GeForce GTX 480" },
	{ 0x10DE06C3,	"NVIDIA GeForce GTX D12U" },
	{ 0x10DE06C4,	"NVIDIA GeForce GTX 465" },
	{ 0x10DE06CA,	"NVIDIA GeForce GTX 480M" },
	{ 0x10DE06CD,	"NVIDIA GeForce GTX 470" },
	// 06D0 - 06DF
	{ 0x10DE06D1,	"NVIDIA Tesla C2050" },
	{ 0x10DE06D2,	"NVIDIA Tesla M2070" },
	{ 0x10DE06D8,	"NVIDIA Quadro 6000" },
	{ 0x10DE06D9,	"NVIDIA Quadro 5000" },
	{ 0x10DE06DA,	"NVIDIA Quadro 5000M" },
	{ 0x10DE06DC,	"NVIDIA Quadro 6000" },
	{ 0x10DE06DD,	"NVIDIA Quadro 4000" },
	{ 0x10DE06DE,	"NVIDIA Tesla M2050" },
	{ 0x10DE06DF,	"NVIDIA Tesla M2070-Q" },
	// 06E0 - 06EF
	{ 0x10DE06E0,	"NVIDIA GeForce 9300 GE" },
	{ 0x10DE06E1,	"NVIDIA GeForce 9300 GS" },
	{ 0x10DE06E2,	"NVIDIA GeForce 8400" },
	{ 0x10DE06E3,	"NVIDIA GeForce 8400 SE" },
	{ 0x10DE06E4,	"NVIDIA GeForce 8400 GS" },
	{ 0x10DE06E5,	"NVIDIA GeForce 9300M GS" },
	{ 0x10DE06E6,	"NVIDIA GeForce G100" },
	{ 0x10DE06E7,	"NVIDIA GeForce 9300 SE" },
	{ 0x10DE06E8,	"NVIDIA GeForce 9200M GE" },
	{ 0x10DE06E9,	"NVIDIA GeForce 9300M GS" },
	{ 0x10DE06EA,	"NVIDIA Quadro NVS 150M" },
	{ 0x10DE06EB,	"NVIDIA Quadro NVS 160M" },
	{ 0x10DE06EC,	"NVIDIA GeForce G 105M" },
	{ 0x10DE06ED,	"NVIDIA G98" },
	{ 0x10DE06EF,	"NVIDIA GeForce G 103M" },
	// 06F0 - 06FF
	{ 0x10DE06F0,	"NVIDIA G98" },
	{ 0x10DE06F1,	"NVIDIA GeForce G105M" },
	{ 0x10DE06F2,	"NVIDIA G98" },
	{ 0x10DE06F3,	"NVIDIA G98" },
	{ 0x10DE06F4,	"NVIDIA G98" },
	{ 0x10DE06F5,	"NVIDIA G98" },
	{ 0x10DE06F6,	"NVIDIA G98" },
	{ 0x10DE06F7,	"NVIDIA G98" },
	{ 0x10DE06F8,	"NVIDIA Quadro NVS 420" },
	{ 0x10DE06F9,	"NVIDIA Quadro FX 370 LP" },
	{ 0x10DE06FA,	"NVIDIA Quadro NVS 450" },
	{ 0x10DE06FB,	"NVIDIA Quadro FX 370M" },
	{ 0x10DE06FC,	"NVIDIA G98" },
	{ 0x10DE06FD,	"NVIDIA Quadro NVS 295" },
	{ 0x10DE06FE,	"NVIDIA G98" },
	{ 0x10DE06FF,	"HICx16 + Graphics" },
	// 0700 - 070F
	// 0710 - 071F
	// 0720 - 072F
	// 0730 - 073F
	// 0740 - 074F
	// 0750 - 075F
	// 0760 - 076F
	// 0770 - 077F
	// 0780 - 078F
	// 0790 - 079F
	// 07A0 - 07AF
	// 07B0 - 07BF
	// 07C0 - 07CF
	// 07D0 - 07DF
	// 07E0 - 07EF
	{ 0x10DE07E0,	"NVIDIA GeForce 7150 / nForce 630i" },
	{ 0x10DE07E1,	"NVIDIA GeForce 7100 / nForce 630i" },
	{ 0x10DE07E2,	"NVIDIA GeForce 7050 / nForce 630i" },
	{ 0x10DE07E3,	"NVIDIA GeForce 7050 / nForce 610i" },
	{ 0x10DE07E5,	"NVIDIA GeForce 7050 / nForce 620i" },
	// 07F0 - 07FF
	// 0800 - 080F
	// 0810 - 081F
	// 0820 - 082F
	// 0830 - 083F
	// 0840 - 084F
	{ 0x10DE0840,	"NVIDIA GeForce 8200M" },
	{ 0x10DE0844,	"NVIDIA GeForce 9100M G" },
	{ 0x10DE0845,	"NVIDIA GeForce 8200M G" },
	{ 0x10DE0846,	"NVIDIA GeForce 9200" },
	{ 0x10DE0847,	"NVIDIA GeForce 9100" },
	{ 0x10DE0848,	"NVIDIA GeForce 8300" },
	{ 0x10DE0849,	"NVIDIA GeForce 8200" },
	{ 0x10DE084A,	"nForce 730a" },
	{ 0x10DE084B,	"NVIDIA GeForce 9200" },
	{ 0x10DE084C,	"nForce 980a/780a SLI" },
	{ 0x10DE084D,	"nForce 750a SLI" },
	{ 0x10DE084F,	"NVIDIA GeForce 8100 / nForce 720a" },
	// 0850 - 085F
	// 0860 - 086F
	{ 0x10DE0860,	"NVIDIA GeForce 9300" },
	{ 0x10DE0861,	"NVIDIA GeForce 9400" },
	{ 0x10DE0862,	"NVIDIA GeForce 9400M G" },
	{ 0x10DE0863,	"NVIDIA GeForce 9400M" },
	{ 0x10DE0864,	"NVIDIA GeForce 9300" },
	{ 0x10DE0865,	"NVIDIA GeForce 9300" },
	{ 0x10DE0866,	"NVIDIA GeForce 9400M G" },
	{ 0x10DE0867,	"NVIDIA GeForce 9400" },
	{ 0x10DE0868,	"nForce 760i SLI" },
	{ 0x10DE0869,	"NVIDIA GeForce 9400" },
	{ 0x10DE086A,	"NVIDIA GeForce 9400" },
	{ 0x10DE086C,	"NVIDIA GeForce 9300 / nForce 730i" },
	{ 0x10DE086D,	"NVIDIA GeForce 9200" },
	{ 0x10DE086E,	"NVIDIA GeForce 9100M G" },
	{ 0x10DE086F,	"NVIDIA GeForce 8200M G" },
	// 0870 - 087F
	{ 0x10DE0870,	"NVIDIA GeForce 9400M" },
	{ 0x10DE0871,	"NVIDIA GeForce 9200" },
	{ 0x10DE0872,	"NVIDIA GeForce G102M" },
	{ 0x10DE0873,	"NVIDIA GeForce G205M" },
	{ 0x10DE0874,	"ION 9300M" },
	{ 0x10DE0876,	"ION 9400M" },
	{ 0x10DE087A,	"NVIDIA GeForce 9400" },
	{ 0x10DE087D,	"ION 9400M" },
	{ 0x10DE087E,	"ION LE" },
	{ 0x10DE087F,	"ION LE" }, // Tesla M2070-Q ??
	// 0880 - 088F
	// 0890 - 089F
	// 08A0 - 08AF
	{ 0x10DE08A0,	"NVIDIA GeForce 320M" },
	{ 0x10DE08A1,	"MCP89-MZT" },
	{ 0x10DE08A2,	"NVIDIA GeForce 320M" },
	{ 0x10DE08A3,	"NVIDIA GeForce 320M" },
	{ 0x10DE08A4,	"NVIDIA GeForce 320M" },
	{ 0x10DE08A5,	"NVIDIA GeForce 320M" },
	// 08B0 - 08BF
	{ 0x10DE08B0,	"MCP83 MMD" },
	{ 0x10DE08B1,	"NVIDIA GeForce 300M" },
	{ 0x10DE08B2,	"NVIDIA GeForce 300M" }, // MCP83-MJ
	{ 0x10DE08B3,	"MCP89 MM9" },
	// 08C0 - 08CF
	// 08D0 - 08DF
	// 08E0 - 08EF
	// 08F0 - 08FF
	// 0900 - 090F
	// 0910 - 091F
	// 0920 - 092F
	// 0930 - 093F
	// 0940 - 094F
	// 0950 - 095F
	// 0960 - 096F
	// 0970 - 097F
	// 0980 - 098F
	// 0990 - 099F
	// 09A0 - 09AF
	// 09B0 - 09BF
	// 09C0 - 09CF
	// 09D0 - 09DF
	// 09E0 - 09EF
	// 09F0 - 09FF
	// 0A00 - 0A0F
	// { 0x10DE0A00,	"NVIDIA GT212" },
	// 0A10 - 0A1F
	// { 0x10DE0A10,	"NVIDIA GT212" },
	// 0A20 - 0A2F
	{ 0x10DE0A20,	"NVIDIA GeForce GT 220" },
	{ 0x10DE0A21,	"D10M2-20" },
	{ 0x10DE0A22,	"NVIDIA GeForce 315" },
	{ 0x10DE0A23,	"NVIDIA GeForce 210" },
	{ 0x10DE0A26,	"NVIDIA GeForce 405" },
	{ 0x10DE0A27,	"NVIDIA GeForce 405" },
	{ 0x10DE0A28,	"NVIDIA GeForce GT 230" },
	{ 0x10DE0A29,	"NVIDIA GeForce GT 330M" },
	{ 0x10DE0A2A,	"NVIDIA GeForce GT 230M" },
	{ 0x10DE0A2B,	"NVIDIA GeForce GT 330M" },
	{ 0x10DE0A2C,	"NVIDIA NVS 5100M" },
	{ 0x10DE0A2D,	"NVIDIA GeForce GT 320M" },	
	// 0A30 - 0A3F
	{ 0x10DE0A30,	"NVIDIA GeForce GT 330M" },
	{ 0x10DE0A32,	"NVIDIA GeForce GT 415" },
	{ 0x10DE0A34,	"NVIDIA GeForce GT 240M" },
	{ 0x10DE0A35,	"NVIDIA GeForce GT 325M" },
	{ 0x10DE0A38,	"NVIDIA Quadro 400" },
	{ 0x10DE0A3C,	"NVIDIA Quadro FX 880M" },
	{ 0x10DE0A3D,	"N10P-ES" },
	{ 0x10DE0A3F,	"GT216-INT" },
	// 0A40 - 0A4F
	// 0A50 - 0A5F
	// 0A60 - 0A6F
	{ 0x10DE0A60,	"NVIDIA GeForce G210" },
	{ 0x10DE0A61,	"NVIDIA NVS 2100" },
	{ 0x10DE0A62,	"NVIDIA GeForce 205" },
	{ 0x10DE0A63,	"NVIDIA GeForce 310" },
	{ 0x10DE0A64,	"ION" },
	{ 0x10DE0A65,	"NVIDIA GeForce 210" },
	{ 0x10DE0A66,	"NVIDIA GeForce 310" },
	{ 0x10DE0A67,	"NVIDIA GeForce 315" },
	{ 0x10DE0A68,	"NVIDIA GeForce G105M" },
	{ 0x10DE0A69,	"NVIDIA GeForce G105M" },
	{ 0x10DE0A6A,	"NVIDIA NVS 2100M" },
	{ 0x10DE0A6C,	"NVIDIA NVS 3100M" },
	{ 0x10DE0A6E,	"NVIDIA GeForce 305M" },
	{ 0x10DE0A6F,	"ION" },	
	// 0A70 - 0A7F
	{ 0x10DE0A70,	"NVIDIA GeForce 310M" },
	{ 0x10DE0A71,	"NVIDIA GeForce 305M" },
	{ 0x10DE0A72,	"NVIDIA GeForce 310M" },
	{ 0x10DE0A73,	"NVIDIA GeForce 305M" },
	{ 0x10DE0A74,	"NVIDIA GeForce G210M" },
	{ 0x10DE0A75,	"NVIDIA GeForce G310M" },
	{ 0x10DE0A76,	"ION" },
	{ 0x10DE0A78,	"NVIDIA Quadro FX 380 LP" },
	// { 0x10DE0A79,	"N12M-NS-S" },
	{ 0x10DE0A7A,	"NVIDIA GeForce 315M" },
	{ 0x10DE0A7B,	"NVIDIA GeForce 505" },
	{ 0x10DE0A7C,	"NVIDIA Quadro FX 380M" },
	{ 0x10DE0A7D,	"N11M-ES" }, //SUBIDS
	{ 0x10DE0A7E,	"GT218-INT-S" },
	{ 0x10DE0A7F,	"GT218-INT-B" },
	// 0A80 - 0A8F
	// 0A90 - 0A9F
	// 0AA0 - 0AAF
	// 0AB0 - 0ABF
	// 0AC0 - 0ACF
	// 0AD0 - 0ADF
	// 0AE0 - 0AEF
	// 0AF0 - 0AFF
	// 0B00 - 0B0F
	// 0B10 - 0B1F
	// 0B20 - 0B2F
	// 0B30 - 0B3F
	// 0B40 - 0B4F
	// 0B50 - 0B5F
	// 0B60 - 0B6F
	// 0B70 - 0B7F
	// 0B80 - 0B8F
	// 0B90 - 0B9F
	// 0BA0 - 0BAF
	// 0BB0 - 0BBF
	// 0BC0 - 0BCF
	// 0BD0 - 0BDF
	// 0BE0 - 0BEF
	// 0BF0 - 0BFF
	// 0C00 - 0C0F
	// 0C10 - 0C1F
	// 0C20 - 0C2F
	// 0C30 - 0C3F
	// 0C40 - 0C4F
	// 0C50 - 0C5F
	// 0C60 - 0C6F
	// 0C70 - 0C7F
	// 0C80 - 0C8F
	// 0C90 - 0C9F
	// 0CA0 - 0CAF
	{ 0x10DE0CA0,	"NVIDIA GeForce GT 330 " },
	{ 0x10DE0CA2,	"NVIDIA GeForce GT 320" },
	{ 0x10DE0CA3,	"NVIDIA GeForce GT 240" },
	{ 0x10DE0CA4,	"NVIDIA GeForce GT 340" },
	{ 0x10DE0CA5,	"NVIDIA GeForce GT 220" },
	{ 0x10DE0CA7,	"NVIDIA GeForce GT 330" },
	{ 0x10DE0CA8,	"NVIDIA GeForce GTS 260M" },
	{ 0x10DE0CA9,	"NVIDIA GeForce GTS 250M" },
	{ 0x10DE0CAC,	"NVIDIA GeForce GT 220" },
	{ 0x10DE0CAD,	"N10E-ES" }, // SUBIDS
	{ 0x10DE0CAE,	"GT215-INT" },
	{ 0x10DE0CAF,	"NVIDIA GeForce GT 335M" },
	// 0CB0 - 0CBF	
	{ 0x10DE0CB0,	"NVIDIA GeForce GTS 350M" },
	{ 0x10DE0CB1,	"NVIDIA GeForce GTS 360M" },
	{ 0x10DE0CBC,	"NVIDIA Quadro FX 1800M" },
	// 0CC0 - 0CCF
	// 0CD0 - 0CDF
	// 0CE0 - 0CEF
	// 0CF0 - 0CFF
	// 0D00 - 0D0F
	// 0D10 - 0D1F
	// 0D20 - 0D2F
	// 0D30 - 0D3F
	// 0D40 - 0D4F
	// 0D50 - 0D5F
	// 0D60 - 0D6F
	// 0D70 - 0D7F
	// 0D80 - 0D8F
	// 0D90 - 0D9F
	// 0DA0 - 0DAF
	// 0DB0 - 0DBF
	// 0DC0 - 0DCF
	{ 0x10DE0DC0,	"NVIDIA GeForce GT 440" },
	//  { 0x10DE0DC1,	"D12-P1-35" },
	//  { 0x10DE0DC2,	"D12-P1-35" },
	{ 0x10DE0DC4,	"NVIDIA GeForce GTS 450" },
	{ 0x10DE0DC5,	"NVIDIA GeForce GTS 450" },
	{ 0x10DE0DC6,	"NVIDIA GeForce GTS 450" },
	//  { 0x10DE0DCA,	"GF10x" },
	//  { 0x10DE0DCC,	"N12E-GS" },
	{ 0x10DE0DCD,	"NVIDIA GeForce GT 555M" },
	{ 0x10DE0DCE,	"NVIDIA GeForce GT 555M" },
	//  { 0x10DE0DCF,	"N12P-GT-B" },
	// 0DD0 - 0DDF	
	//  { 0x10DE0DD0,	"N11E-GT" },
	{ 0x10DE0DD1,	"NVIDIA GeForce GTX 460M" },
	{ 0x10DE0DD2,	"NVIDIA GeForce GT 445M" },
	{ 0x10DE0DD3,	"NVIDIA GeForce GT 435M" },
	{ 0x10DE0DD6,	"NVIDIA GeForce GT 550M" },
	{ 0x10DE0DD8,	"NVIDIA Quadro 2000" },
	{ 0x10DE0DDA,	"NVIDIA Quadro 2000M" },
	{ 0x10DE0DDE,	"GF106-ES" },
	//  { 0x10DE0DDF,	"GF106-INT" },
	// 0DE0 - 0DEF
	{ 0x10DE0DE0,	"NVIDIA GeForce GT 440" },
	{ 0x10DE0DE1,	"NVIDIA GeForce GT 430" },
	{ 0x10DE0DE2,	"NVIDIA GeForce GT 420" },
	{ 0x10DE0DE3,	"NVIDIA GeForce GT 635M" },
	{ 0x10DE0DE4,	"NVIDIA GeForce GT 520" },
	{ 0x10DE0DE5,	"NVIDIA GeForce GT 530" },
	{ 0x10DE0DE8,	"NVIDIA GeForce GT 620M" },
	{ 0x10DE0DE9,	"NVIDIA GeForce GT 630M" },
	{ 0x10DE0DEA,	"NVIDIA GeForce GT 610M" },
	{ 0x10DE0DEB,	"NVIDIA GeForce GT 555M" },
	{ 0x10DE0DEC,	"NVIDIA GeForce GT 525M" },
	{ 0x10DE0DED,	"NVIDIA GeForce GT 520M" },
	{ 0x10DE0DEE,	"NVIDIA GeForce GT 415M" },
	{ 0x10DE0DEF,	"N13P-NS1-A1" },
	// 0DF0 - 0DFF
	{ 0x10DE0DF0,	"NVIDIA GeForce GT 425M" },
	{ 0x10DE0DF1,	"NVIDIA GeForce GT 420M" },
	{ 0x10DE0DF2,	"NVIDIA GeForce GT 435M" },
	{ 0x10DE0DF3,	"NVIDIA GeForce GT 420M" },
	{ 0x10DE0DF4,	"NVIDIA GeForce GT 540M" },
	{ 0x10DE0DF5,	"NVIDIA GeForce GT 525M" },
	{ 0x10DE0DF6,	"NVIDIA GeForce GT 550M" },
	{ 0x10DE0DF7,	"NVIDIA GeForce GT 520M" },
	{ 0x10DE0DF8,	"NVIDIA Quadro 600" },
	{ 0x10DE0DF9,	"NVIDIA Quadro 500M" },
	{ 0x10DE0DFA,	"NVIDIA Quadro 1000M" },
	{ 0x10DE0DFC,	"NVIDIA NVS 5200M" },
	{ 0x10DE0DFE,	"GF108 ES" },
	//  { 0x10DE0DFF,	"GF108 INT" },
	// 0E00 - 0E0F
	// 0E10 - 0E1F
	// 0E20 - 0E2F
	{ 0x10DE0E21,	"D12U-25" },
	{ 0x10DE0E22,	"NVIDIA GeForce GTX 460" },
	{ 0x10DE0E23,	"NVIDIA GeForce GTX 460 SE" },
	{ 0x10DE0E24,	"NVIDIA GeForce GTX 460" },
	//  { 0x10DE0E25,	"D12U-50" },
	{ 0x10DE0E28,	"NVIDIA GeForce GTX 460" },
	// 0E30 - 0E3F
	{ 0x10DE0E30,	"NVIDIA GeForce GTX 470M" },
	{ 0x10DE0E31,	"NVIDIA GeForce GTX 485M" },
	//  { 0x10DE0E32,	"N12E-GT" },
	{ 0x10DE0E38,	"GF104GL" },
	{ 0x10DE0E3A,	"NVIDIA Quadro 3000M" },
	{ 0x10DE0E3B,	"NVIDIA Quadro 4000M" },
	//  { 0x10DE0E3E,	"GF104-ES" },
	//  { 0x10DE0E3F,	"GF104-INT" },
	// 0E40 - 0E4F
	// 0E50 - 0E5F
	// 0E60 - 0E6F
	// 0E70 - 0E7F
	// 0E80 - 0E8F
	// 0E90 - 0E9F
	// 0EA0 - 0EAF
	// 0EB0 - 0EBF
	// 0EC0 - 0ECF
	// 0ED0 - 0EDF
	// 0EE0 - 0EEF
	// 0EF0 - 0EFF
	// 0F00 - 0F0F
	{ 0x10DE0F00,	"NVIDIA GeForce GT 630" },
	{ 0x10DE0F01,	"NVIDIA GeForce GT 620" },
	// 0F10 - 0F1F
	// 0F20 - 0F2F
	// 0F30 - 0F3F
	// 0F40 - 0F4F
	// 0F50 - 0F5F
	// 0F60 - 0F6F
	// 0F70 - 0F7F
	// 0F80 - 0F8F
	// 0F90 - 0F9F
	// 0FA0 - 0FAF
	// 0FB0 - 0FBF
	// 0FC0 - 0FCF
	{ 0x10DE0FC0,	"NVIDIA GeForce GT 640" },
	{ 0x10DE0FC1,	"NVIDIA GeForce GT 640" },
	{ 0x10DE0FC2,	"NVIDIA GeForce GT 630" },
	{ 0x10DE0FC6,	"NVIDIA GeForce GTX 650" },
	{ 0x10DE0FCD,	"NVIDIA GeForce GT 755M" },
	// 0FD0 - 0FDF
	{ 0x10DE0FD1,	"NVIDIA GeForce GT 650M" },
	{ 0x10DE0FD2,	"NVIDIA GeForce GT 640M" },
	{ 0x10DE0FD3,	"NVIDIA GeForce GT 640M LE" },
	{ 0x10DE0FD4,	"NVIDIA GeForce GTX 660M" },
	{ 0x10DE0FD5,	"NVIDIA GeForce GT 650M" },
	{ 0x10DE0FD8,	"NVIDIA GeForce GT 640M" },
	{ 0x10DE0FD9,	"NVIDIA GeForce GT 645M" },
	{ 0x10DE0FDB,	"GK107-ESP-A1" },
	{ 0x10DE0FDF,	"NVIDIA GeForce GT 740M" },
	// 0FE0 - 0FEF
	{ 0x10DE0FE0,	"NVIDIA GeForce GTX 660M" },
	{ 0x10DE0FE1,	"NVIDIA GeForce GT 730M" },
	{ 0x10DE0FE3,	"NVIDIA GeForce GT 745M" },
	{ 0x10DE0FE4,	"NVIDIA GeForce GT 750M" },
	// 0FF0 - 0FFF
	{ 0x10DE0FF2,	"VGX K1" },
	{ 0x10DE0FF8,	"NVIDIA Quadro K500M" },
	{ 0x10DE0FF9,	"NVIDIA Quadro K2000D" },
	{ 0x10DE0FFA,	"NVIDIA Quadro K600" },
	{ 0x10DE0FFB,	"NVIDIA Quadro K2000M" },
	{ 0x10DE0FFC,	"NVIDIA Quadro K1000M" },
	{ 0x10DE0FFD,	"NVIDIA NVS 510" },
	{ 0x10DE0FFE,	"NVIDIA Quadro K2000" },
	{ 0x10DE0FFF,	"NVIDIA Quadro 410" },
	// 1000 - 100F
	{ 0x10DE1003,	"NVIDIA GeForce GTX Titan LE" },
	{ 0x10DE1004,	"NVIDIA GeForce GTX 780" },
	{ 0x10DE1005,	"NVIDIA GeForce GTX Titan" },
	{ 0x10DE100A,	"NVIDIA GeForce GTX 780 Ti" },
	// 1010 - 101F
	{ 0x10DE101F,	"NVIDIA Tesla K20" },
	// 1020 - 102F
	{ 0x10DE1020,	"NVIDIA Tesla K20X" },
	{ 0x10DE1021,	"NVIDIA Tesla K20Xm" },
	{ 0x10DE1022,	"NVIDIA Tesla K20c" },
	{ 0x10DE1026,	"NVIDIA Tesla K20s" },
	{ 0x10DE1028,	"NVIDIA Tesla K20m" },
	// 1030 - 103F
	// 1040 - 104F
	{ 0x10DE1040,	"NVIDIA GeForce GT 520" },
	//  { 0x10DE1041,	"D13M1-45" },
	{ 0x10DE1042,	"NVIDIA GeForce 510" },
	{ 0x10DE1048,	"NVIDIA GeForce 605" },
	{ 0x10DE1049,	"NVIDIA GeForce GT 620" },
	{ 0x10DE104A,	"NVIDIA GeForce GT 610" },
	{ 0x10DE104B,	"NVIDIA GeForce GT 625 (OEM)" },
	// 1050 - 105F
	{ 0x10DE1050,	"NVIDIA GeForce GT 520M" },
	{ 0x10DE1051,	"NVIDIA GeForce GT 520MX" },
	{ 0x10DE1052,	"NVIDIA GeForce GT 520M" },
	{ 0x10DE1054,	"NVIDIA GeForce GT 410M" },
	{ 0x10DE1055,	"NVIDIA GeForce 410M" },
	{ 0x10DE1056,	"NVIDIA Quadro NVS 4200M" },
	{ 0x10DE1057,	"NVIDIA Quadro NVS 4200M" },
	{ 0x10DE1058,	"NVIDIA GeForce GT 610M" },
	{ 0x10DE1059,	"NVIDIA GeForce 610M" },
	{ 0x10DE105A,	"NVIDIA GeForce 610M" },
	{ 0x10DE105B,	"NVIDIA GeForce 705A" },
	// 1060 - 106F
	// 1070 - 107F
	{ 0x10DE107C,	"NVIDIA Quadro NVS 315" },
	{ 0x10DE107D,	"NVIDIA Quadro NVS 310" },
	//  { 0x10DE107E,	"GF119-INT" },
	{ 0x10DE107F,	"GF119-ES" },
	// 1080 - 108F
	{ 0x10DE1080,	"NVIDIA GeForce GTX 580" },
	{ 0x10DE1081,	"NVIDIA GeForce GTX 570" },
	{ 0x10DE1082,	"NVIDIA GeForce GTX 560 Ti" },
	{ 0x10DE1083,	"D13U" },
	{ 0x10DE1084,	"NVIDIA GeForce GTX 560" },
	{ 0x10DE1086,	"NVIDIA GeForce GTX 570 HD" },
	{ 0x10DE1087,	"NVIDIA GeForce GTX 560 Ti-448" },
	{ 0x10DE1088,	"NVIDIA GeForce GTX 590" },
	{ 0x10DE1089,	"NVIDIA GeForce GTX 580" },
	{ 0x10DE108B,	"NVIDIA GeForce GTX 590" },
	//  { 0x10DE108C,	"D13U" },
	{ 0x10DE108E,	"NVIDIA Tesla C2090" },
	// 1090 - 109F
	{ 0x10DE1091,	"NVIDIA Tesla M2090" }, // X2090
	{ 0x10DE1094,	"NVIDIA Tesla M2075" },
	{ 0x10DE1096,	"NVIDIA Tesla C2075" },
	{ 0x10DE1098,	"D13U" },
	{ 0x10DE109A,	"NVIDIA Quadro 5010M" },
	{ 0x10DE109B,	"NVIDIA Quadro 7000" },
	// 10A0 - 10AF
	// 10B0 - 10BF
	// 10C0 - 10CF
	{ 0x10DE10C0,	"NVIDIA GeForce 9300 GS" },
	{ 0x10DE10C3,	"NVIDIA GeForce 8400 GS" },
	{ 0x10DE10C4,	"ION" },
	{ 0x10DE10C5,	"NVIDIA GeForce 405" },
	// 10D0 - 10DF
	{ 0x10DE10D8,	"NVIDIA Quadro NVS 300" },
	// 10E0 - 10EF
	// 10F0 - 10FF
	// 1100 - 110F
	// 1110 - 111F
	// 1120 - 112F
	// 1130 - 113F
	// 1140 - 114F
	{ 0x10DE1140,	"NVIDIA GeForce GT 610M" },
	{ 0x10DE1141,	"NVIDIA GeForce 610M" },
	{ 0x10DE1142,	"NVIDIA GeForce 620M" },
	//  { 0x10DE1143,	"N13P-GV" },
	//  { 0x10DE1144,	"GF117" },
	//  { 0x10DE1145,	"GF117" },
	//  { 0x10DE1146,	"GF117" },
	//  { 0x10DE1147,	"GF117" },
	{ 0x10DE1149,	"GF117-ES" },
	//  { 0x10DE114A,	"GF117-INT" },
	//  { 0x10DE114B,	"PCI-GEN3-B" },
	// 1150 - 115F
	// 1160 - 116F
	// 1170 - 117F
	// 1180 - 118F
	{ 0x10DE1180,	"NVIDIA GeForce GTX 680" },
	{ 0x10DE1182,	"NVIDIA GeForce GTX 760 Ti" },
	{ 0x10DE1183,	"NVIDIA GeForce GTX 660 Ti" },
	{ 0x10DE1184,	"NVIDIA GeForce GTX 770" },
	{ 0x10DE1185,	"NVIDIA GeForce GTX 660" },
	{ 0x10DE1187,	"NVIDIA GeForce GTX 760" },
	{ 0x10DE1188,	"NVIDIA GeForce GTX 690" },
	{ 0x10DE1189,	"NVIDIA GeForce GTX 670" },
	{ 0x10DE118A,	"GRID K520" },
	{ 0x10DE118B,	"GRID K200" }, // GRID K2 GeForce USM
	{ 0x10DE118E,	"NVIDIA GeForce GTX 760 (192-bit)" },
	{ 0x10DE118F,	"NVIDIA Tesla K10" },
	// 1190 - 119F
	{ 0x10DE1192,	"NVIDIA GeForce GK104" },
	{ 0x10DE1193,	"NVIDIA GeForce GTX 760 Ti" },
	{ 0x10DE1198,	"NVIDIA GeForce GTX 880M" },
	{ 0x10DE1199,	"NVIDIA GeForce GTX 870M" },
	{ 0x10DE119a,	"NVIDIA GeForce GTX 860M" },
	{ 0x10DE119d,	"NVIDIA GeForce GTX 775M" }, // Mac Edition
	{ 0x10DE119e,	"NVIDIA GeForce GTX 780M" }, // Mac Edition
	{ 0x10DE119F,	"NVIDIA GeForce GTX 780M" },
	// 11A0 - 11AF
	{ 0x10DE11A0,	"NVIDIA GeForce GTX 680M" },
	{ 0x10DE11A1,	"NVIDIA GeForce GTX 670MX" },
	{ 0x10DE11A2,	"NVIDIA GeForce GTX 675MX" }, // Mac Edition
	{ 0x10DE11A3,	"NVIDIA GeForce GTX 680MX" },
	{ 0x10DE11A7,	"NVIDIA GeForce GTX 675MX" },
	// 11B0 - 11BF
	{ 0x10DE11B0,	"GRID K240Q" }, // K260Q vGPU
	{ 0x10DE11B1,	"GRID K2 Tesla USM" },
	{ 0x10DE11B6,	"NVIDIA Quadro K3100M" },
	{ 0x10DE11B7,	"NVIDIA Quadro K4100M" },
	{ 0x10DE11B8,	"NVIDIA Quadro K5100M" },
	{ 0x10DE11BA,	"NVIDIA Quadro K5000" },
	{ 0x10DE11BB,	"NVIDIA Quadro 4100" },
	{ 0x10DE11BC,	"NVIDIA Quadro K5000M" },
	{ 0x10DE11BD,	"NVIDIA Quadro K4000M" },
	{ 0x10DE11BE,	"NVIDIA Quadro K3000M" },
	{ 0x10DE11BF,	"VGX K2" },
	// 11C0 - 11CF
	{ 0x10DE11C0,	"NVIDIA GeForce GTX 660" },
	{ 0x10DE11C2,	"NVIDIA GeForce GTX 650 Ti BOOST" },
	{ 0x10DE11C3,	"NVIDIA GeForce GTX 650 Ti" },
	{ 0x10DE11C4,	"NVIDIA GeForce GTX 645" },
	{ 0x10DE11C6,	"NVIDIA GeForce GTX 650 Ti" },
	{ 0x10DE11C7,	"NVIDIA GeForce GTX 750 Ti" },
	{ 0x10DE11C8,	"NVIDIA GeForce GTX 650 OEM" },
	// 11D0 - 11DF
	{ 0x10DE11D0,	"GK106-INT353" },
	// 11E0 - 11EF
	{ 0x10DE11E0,	"NVIDIA GeForce GTX 770M" },
	{ 0x10DE11E1,	"N14E-GE-B-A1" },
	{ 0x10DE11E2,	"NVIDIA GeForce GTX 765M" },
	{ 0x10DE11E3,	"NVIDIA GeForce GTX 760M" },
	// 11F0 - 11FF
	{ 0x10DE11FA,	"NVIDIA Quadro K4000" },
	{ 0x10DE11FC,	"NVIDIA Quadro 2100M" },
	// 1200 - 120F
	{ 0x10DE1200,	"NVIDIA GeForce GTX 560 Ti" },
	{ 0x10DE1201,	"NVIDIA GeForce GTX 560" },
	{ 0x10DE1202,	"NVIDIA GeForce GTX 560 Ti" },
	{ 0x10DE1203,	"NVIDIA GeForce GTX 460 SE v2" },
	{ 0x10DE1205,	"NVIDIA GeForce GTX 460 v2" },
	{ 0x10DE1206,	"NVIDIA GeForce GTX 555" },
	{ 0x10DE1207,	"NVIDIA GeForce GT 645" },
	{ 0x10DE1208,	"NVIDIA GeForce GTX 560 SE" },
	{ 0x10DE1210,	"NVIDIA GeForce GTX 570M" },
	{ 0x10DE1211,	"NVIDIA GeForce GTX 580M" },
	{ 0x10DE1212,	"NVIDIA GeForce GTX 675M" },
	{ 0x10DE1213,	"NVIDIA GeForce GTX 670M" },
	//{ 0x10DE121F,	"GF114-INT" },
	{ 0x10DE1240,	"NVIDIA GeForce GT 620M" },
	{ 0x10DE1241,	"NVIDIA GeForce GT 545" },
	{ 0x10DE1243,	"NVIDIA GeForce GT 545" },
	{ 0x10DE1244,	"NVIDIA GeForce GTX 550 Ti" },
	{ 0x10DE1245,	"NVIDIA GeForce GTS 450" },
	{ 0x10DE1246,	"NVIDIA GeForce GT 550M" },
	{ 0x10DE1247,	"NVIDIA GeForce GT 555M" },
	{ 0x10DE1248,	"NVIDIA GeForce GT 555M" },
	{ 0x10DE1249,	"NVIDIA GeForce GTS 450" },
	{ 0x10DE124B,	"NVIDIA GeForce GT 640" },
	{ 0x10DE124D,	"NVIDIA GeForce GT 555M" },
	{ 0x10DE1250,	"GF116-INT" },
	{ 0x10DE1251,	"NVIDIA GeForce GTX 560M" },
	// 1260 - 126F
	// 1270 - 127F
	// 1280 - 128F
	{ 0x10DE1280,	"NVIDIA GeForce GT 635" },
	{ 0x10DE1281,	"NVIDIA GeForce GT 710" },
	{ 0x10DE1282,	"NVIDIA GeForce GT 640" },
	{ 0x10DE1284,	"NVIDIA GeForce GT 630" },
	// 1290 - 129F
	{ 0x10DE1290,	"NVIDIA GeForce GT 730M" },
	{ 0x10DE1291,	"NVIDIA GeForce GT 735M" },
	{ 0x10DE1292,	"NVIDIA GeForce GT 740M" },
	{ 0x10DE1293,	"NVIDIA GeForce GT 730M" },
	{ 0x10DE1294,	"NVIDIA GeForce GT 740M" },
	{ 0x10DE1295,	"NVIDIA GeForce GT 710M" },
	{ 0x10DE1296,	"NVIDIA GeForce 825M" }, // GK208M
	{ 0x10DE1298,	"NVIDIA GeForce GT 720M" },
	// 12A0 - 12AF
	//{ 0x10DE12A0,	"NVIDIA GeForce GT ???" },
	{ 0x10DE12AF,	"GK208-INT" },
	{ 0x10DE12B0,	"GK208-CS-Q" },
	{ 0x10DE12B9,	"NVIDIA Quadro K610M" },
	{ 0x10DE12BA,	"NVIDIA Quadro K510M" },
	// 12B0 - 12BF
	// 12C0 - 12CF
	// 12D0 - 12DF
	// 12E0 - 12EF
	// 12F0 - 12FF
	{ 0x10DE1340,	"NVIDIA GeForce 830M" },
	{ 0x10DE1341,	"NVIDIA GeForce 840M" },
	{ 0x10DE1380,	"NVIDIA GeForce GTX 750 Ti" },
	{ 0x10DE1381,	"NVIDIA GeForce GTX 750" },
	{ 0x10DE1390,	"NVIDIA GeForce 845M" },
	{ 0x10DE1391,	"NVIDIA GeForce GTX 850M" },
	{ 0x10DE1392,	"NVIDIA GeForce GTX 860M" },
	{ 0x10DE1393,	"NVIDIA GeForce 840M" }
};

static nvidia_card_info_t nvidia_card_exceptions[] = {
    /* ========================================================================================
     * Layout is device(VendorId + DeviceId), subdev (SubvendorId + SubdeviceId), display name.
     * ========================================================================================
     */
    /* ------ Specific DeviceID and SubDevID. ------ */
	// 0000 - 00FF
	{ 0x10DE0040,	0x10438178,	"Asus V9999 Ultra V62.11" },
	{ 0x10DE0040,	0x1043817D,	"Asus V9999GT V61.21" },
	{ 0x10DE0040,	0x7FFFFFFF,	"GeForce 6800 Ultra [NV40.0]" },
	// 0100 - 01FF
	{ 0x10DE01D7,	0x102801C2,	"Dell Quadro NVS 110M" },
	{ 0x10DE01D7,	0x102801CC,	"Dell Quadro NVS 110M" },
	{ 0x10DE01D7,	0x10DE014B,	"nVidia Quadro NVS 110M" },

	{ 0x10DE01D8,	0x102801CC,	"Dell Quadro NVS 120M" },
	{ 0x10DE01D8,	0x10282003,	"Dell Quadro NVS 120M" },

	{ 0x10DE01DA,	0x10280407,	"Dell GeForce 7300 LE" },
	// 0200 - 02FF
	{ 0x10DE025B,	0x10480D23,	"ELSA Gloria4 700XGL" },
	// 0300 - 03FF
	{ 0x10DE0391,	0x10DE047A,	"Galaxy GeForce 7600 GT" },
	{ 0x10DE0391,	0x19F120DE,	"Galaxy GeForce 7600 GT" },

	{ 0x10DE0393,	0x00000400,	"Apple GeForce 7300GT" },
	// 0400 - 04FF
	{ 0x10DE0402,	0x10DE0439,	"Galaxy 8600GT" },
	{ 0x10DE0402,	0x10DE0505,	"Galaxy 8600GT" },
	// 0500 - 05FF
	{ 0x10DE05E2,	0x104382EB,	"ASUS ENGTX260" },
	{ 0x10DE05E2,	0x16822390,	"HFX GeForce GTX 260" },
	{ 0x10DE05E2,	0x17870000,	"HIS GeForce GTX 260" },

	{ 0x10DE05E6,	0x10B00401,	"Gainward GeForce GTX 285" },

	{ 0x10DE05E7,	0x10DE0595,	"nVidia Tesla T10 Processor" },
	{ 0x10DE05E7,	0x10DE066A,	"nVidia Tesla C1060" },
	{ 0x10DE05E7,	0x10DE068F,	"nVidia Tesla T10 Processor" },
	{ 0x10DE05E7,	0x10DE0697,	"nVidia Tesla M1060" },
	{ 0x10DE05E7,	0x10DE0714,	"nVidia Tesla M1060" },
	{ 0x10DE05E7,	0x10DE0743,	"nVidia Tesla M1060" },
	// 0600 - 06FF
	{ 0x10DE0600,	0x10DE0000,	"Abit GeForce 8800 GTS" },

	{ 0x10DE0605,	0x145834A2,	"Gigabyte GV-N98TOC-512H" },

	{ 0x10DE0608,	0x15880577,	"Solidum GeForce 9800M GTX" },

	{ 0x10DE0609,	0x11700121,	"Inventec GeForce 8800M GTS" },

	{ 0x10DE0612,	0x104382A6,	"Asus GeForce 9800 GTX+" },
	{ 0x10DE0612,	0x10DE0571,	"nVidia GeForce 9800 GTX+" },
	{ 0x10DE0612,	0x10DE0592,	"nVidia GeForce 9800 GTX+" },
	{ 0x10DE0612,	0x3842C842,	"EVGA GeForce 9800 GTX+" },
	{ 0x10DE0612,	0x3842C875,	"EVGA GeForce 9800 GTX+" },

	{ 0x10DE0615,	0x10480F67,	"ELSA GeForce GTS 250" },
	{ 0x10DE0615,	0x10DE0592,	"Palit GeForce GTS 250" },
	{ 0x10DE0615,	0x10DE0593,	"Palit GeForce GTS 250" },
	{ 0x10DE0615,	0x10DE0652,	"Palit GeForce GTS 250" },
	{ 0x10DE0615,	0x10DE0719,	"Palit GeForce GTS 250" },
	{ 0x10DE0615,	0x10DE079E,	"Palit GeForce GTS 250" },
	{ 0x10DE0615,	0x11503842,	"TMC GeForce GTS 250" }, // Thinking Machines Corporation
	{ 0x10DE0615,	0x11513842,	"JAE GeForce GTS 250" },
	{ 0x10DE0615,	0x11553842,	"Pine GeForce GTS 250" },
	{ 0x10DE0615,	0x11563842,	"Periscope GeForce GTS 250" },

	{ 0x10DE0618,	0x10432028,	"Asus GeForce GTX 170M" },
	{ 0x10DE0618,	0x1043202B,	"Asus GeForce GTX 680" },

	{ 0x10DE0622,	0x104382AC,	"Asus EN9600GT Magic" },

	{ 0x10DE0640,	0x10DE077F,	"Inno3D GeForce 9500GT HDMI" },

	{ 0x10DE0649,	0x1043202D,	"Asus GeForce GT 220M" },

	{ 0x10DE06CD,	0x10DE079F,	"Point of View GeForce GTX 470" },
	{ 0x10DE06CD,	0x14622220,	"MSi GeForce GTX 470 Twin Frozr II" },

	{ 0x10DE06D1,	0x10DE0771,	"nVidia Tesla C2050" },
	{ 0x10DE06D1,	0x10DE0772,	"nVidia Tesla C2070" },

	{ 0x10DE06D2,	0x10DE0774,	"nVidia Tesla M2070" },
	{ 0x10DE06D2,	0x10DE0830,	"nVidia Tesla M2070" },
	{ 0x10DE06D2,	0x10DE0842,	"nVidia Tesla M2070" },
	{ 0x10DE06D2,	0x10DE088F,	"nVidia Tesla X2070" },
	{ 0x10DE06D2,	0x10DE0908,	"nVidia Tesla M2070" },

	{ 0x10DE06DE,	0x10DE0773,	"nVidia Tesla S2050" },
	{ 0x10DE06DE,	0x10DE0830,	"nVidia Tesla M2070" },
	{ 0x10DE06DE,	0x10DE0831,	"nVidia Tesla M2070" },
	{ 0x10DE06DE,	0x10DE0832,	"nVidia Tesla M2070" },
	{ 0x10DE06DE,	0x10DE0840,	"nVidia Tesla X2070" },

	{ 0x10DE06E4,	0x10438322,	"Asus EN8400GS" },
	{ 0x10DE06E4,	0x14583475,	"GV-NX84S256HE [GeForce 8400 GS]" },

	{ 0x10DE06E8,	0x10280262,	"Dell GeForce 9200M GS" },
	{ 0x10DE06E8,	0x10280271,	"Dell GeForce 9200M GS" },
	{ 0x10DE06E8,	0x10280272,	"Dell GeForce 9200M GS" },
	{ 0x10DE06E8,	0x103C30F4,	"HP GeForce 9200M GS" },
	{ 0x10DE06E8,	0x103C30F7,	"HP GeForce 9200M GS" },
	{ 0x10DE06E8,	0x103C3603,	"HP GeForce 9200M GS" },
	// 0700 - 07FF
	// 0800 - 08FF
	{ 0x10DE0873,	0x104319B4,	"Asus GeForce G102M" },
	// 0900 - 09FF
	// 0A00 - 0AFF
	{ 0x10DE0A6F,	0x12974003,	"Shuttle XS 3510MA" },

	{ 0x10DE0A70,	0x17AA3605,	"Lenovo ION" },

	{ 0x10DE0A73,	0x17AA3607,	"Lenovo ION" },
	{ 0x10DE0A73,	0x17AA3610,	"Lenovo ION" },

	{ 0x10DE0A75,	0x17AA3605,	"Lenovo ION" },
	// 0B00 - 0BFF
	// 0C00 - 0CFF
	{ 0x10DE0CA3,	0x14628041,	"MSi VN240GT-MD1G" },
	{ 0x10DE0CA3,	0x16423926,	"Bitland GeForce GT 230" },
	// 0D00 - 0DFF
	{ 0x10DE0DD8,	0x10DE0914,	"nVidia Quadro 2000D" },

	{ 0x10DE0DEF,	0x17AA21F3,	"Lenovo NVS 5400M" },
	{ 0x10DE0DEF,	0x17AA21F4,	"Lenovo NVS 5400M" },
	{ 0x10DE0DEF,	0x17AA21F5,	"Lenovo NVS 5400M" },
	{ 0x10DE0DEF,	0x17AA21F6,	"Lenovo NVS 5400M" },
	{ 0x10DE0DEF,	0x17AA5005,	"Lenovo NVS 5400M" },

	// 0E00 - 0EFF
	{ 0x10DE0E22,	0x1043835D,	"Asus ENGTX460" },

	{ 0x10DE0E23,	0x10B00401,	"Gainward GeForce GTX 460" },
	// 0F00 - 0FFF
	{ 0x10DE0FD2,	0x10280595,	"Dell GeForce GT 640M LE" },
	{ 0x10DE0FD2,	0x102805B2,	"Dell GeForce GT 640M LE" },
	// 1000 - 10FF
	{ 0x10DE1080,	0x14622561,	"MSI N580GTX Lightning" },
	{ 0x10DE1080,	0x14622563,	"MSI N580GTX Lightning" },

	{ 0x10DE1086,	0x10DE0871,	"Inno3D GeForce GTX 570" },

	{ 0x10DE1087,	0x104383D6,	"Asus ENGTX560Ti448 DCII" },

	{ 0x10DE1091,	0x10DE088E,	"nVidia Tesla X2090" },
	{ 0x10DE1091,	0x10DE0891,	"nVidia Tesla X2090" },

	{ 0x10DE1094,	0x10DE0888,	"nVidia Tesla M2075" },

	{ 0x10DE1096,	0x10DE0910,	"nVidia Tesla C2075" },
	{ 0x10DE1096,	0x10DE0911,	"nVidia Tesla C2050" },

	// 1100 - 11FF
	{ 0x10DE1140,	0x1025064A,	"Acer GeForce GT 620M" },
	{ 0x10DE1140,	0x1025064C,	"Acer GeForce GT 620M" },
	{ 0x10DE1140,	0x10250680,	"Acer GeForce GT 620M" },
	{ 0x10DE1140,	0x10250692,	"Acer GeForce GT 620M" },
	{ 0x10DE1140,	0x10250694,	"Acer GeForce GT 620M" },
	{ 0x10DE1140,	0x10250702,	"Acer GeForce GT 620M" },
	{ 0x10DE1140,	0x10250719,	"Acer GeForce GT 620M" },
	{ 0x10DE1140,	0x10250725,	"Acer GeForce GT 620M" },
	{ 0x10DE1140,	0x10250728,	"Acer GeForce GT 620M" },
	{ 0x10DE1140,	0x1025072B,	"Acer GeForce GT 620M" },
	{ 0x10DE1140,	0x1025072E,	"Acer GeForce GT 620M" },
	{ 0x10DE1140,	0x10250732,	"Acer GeForce GT 620M" },
	{ 0x10DE1140,	0x10280565,	"Dell GeForce GT 630M" },
	{ 0x10DE1140,	0x10280568,	"Dell GeForce GT 630M" },
	{ 0x10DE1140,	0x144DC0D5,	"Samsung GeForce GT 630M" },
	{ 0x10DE1140,	0x17AA500D,	"Lenovo GeForce GT 620M" },
	{ 0x10DE1140,	0x1B0A20DD,	"Pegatron GeForce GT 620M" },
	{ 0x10DE1140,	0x1B0A20FD,	"Pegatron GeForce GT 620M" },

	{ 0x10DE1180,	0x00001255,	"Afox GTX 680" },
	{ 0x10DE1180,	0x104383F0,	"Asus GTX680-2GD5" },
	{ 0x10DE1180,	0x104383F6,	"Asus GTX 680 Direct CU II" },
	{ 0x10DE1180,	0x104383F7,	"Asus GTX 680 Direct CU II" },
	{ 0x10DE1180,	0x1458353C,	"GV-N680OC-2GD WindForce GTX 680 OC" },
	{ 0x10DE1180,	0x14622820,	"MSi N680GTX TwinFrozer" },
	{ 0x10DE1180,	0x14622830,	"MSi GTX 680 Lightning" },
	{ 0x10DE1180,	0x14622831,	"MSi GTX 680 Lightning LN2" },
	{ 0x10DE1180,	0x15691180,	"Palit GTX 680 JetStream" },
	{ 0x10DE1180,	0x15691181,	"Palit GTX 680 JetStream" },
	{ 0x10DE1180,	0x15691189,	"Palit GTX 680 JetStream" },
	{ 0x10DE1180,	0x38422682,	"EVGA GTX 680 SC" },
	{ 0x10DE1180,	0x38422683,	"EVGA GTX 680 SC" },

	{ 0x10DE1187,	0x14583614,	"GV-N760OC-4GD" },

	{ 0x10DE1189,	0x10438405,	"Asus GTX 670 Direct CU II TOP" },
	{ 0x10DE1189,	0x15691189,	"Palit GTX 670 JetStream" },
	{ 0x10DE1189,	0x19DA1255,	"Zotac GTX 670 AMP! Edition" },

	{ 0x10DE11A1,	0x15587102,	"Clevo N13E-GR" },

	{ 0x10DE11C0,	0x10DE0995,	"Inno3D GeForce GTX660" },
	{ 0x10DE11C0,	0x1458354E,	"GV-N660OC-2GD" },

	{ 0x10DE11C6,	0x1043842A,	"GTX650TI-1GD5" },
	// 1200 - 12FF
	{ 0x10DE1247,	0x10432119,	"Asus GeForce GT 670M" },
	{ 0x10DE1247,	0x10432120,	"Asus GeForce GT 670M" },
	{ 0x10DE1247,	0x1043212A,	"Asus GeForce GT 635M" },
	{ 0x10DE1247,	0x1043212B,	"Asus GeForce GT 635M" },
	{ 0x10DE1247,	0x1043212C,	"Asus GeForce GT 635M" },
	{ 0x10DE1247,	0x152D0930,	"Quanta GeForce GT 635M" },

	{ 0x10DE1248,	0x152D0930,	"Quanta GeForce GT 635M" },

	{ 0x10DE124D,	0x146210CC,	"MSi GeForce GT 635M" },
};

static int patch_nvidia_rom(uint8_t *rom)
{
	if (!rom || (rom[0] != 0x55 && rom[1] != 0xaa)) {
		printf("False ROM signature: 0x%02x%02x\n", rom[0], rom[1]);
		return PATCH_ROM_FAILED;
	}

	uint16_t dcbptr = READ_LE_SHORT(rom, 0x36);

	if (!dcbptr) {
		printf("no dcb table found\n");
		return PATCH_ROM_FAILED;
	}
//	else
//		printf("dcb table at offset 0x%04x\n", dcbptr);

	uint8_t *dcbtable	 = &rom[dcbptr];
	uint8_t dcbtable_version = dcbtable[0];
	uint8_t headerlength	 = 0;
	uint8_t numentries	 = 0;
	uint8_t recordlength	 = 0;
	
	if (dcbtable_version >= 0x20) {
		uint32_t sig;
		
		if (dcbtable_version >= 0x30) {
			headerlength = dcbtable[1];
			numentries	 = dcbtable[2];
			recordlength = dcbtable[3];

			sig = READ_LE_INT(dcbtable, 6);
		} else {
			sig = READ_LE_INT(dcbtable, 4);
			headerlength = 8;
		}

		if (sig != 0x4edcbdcb) {
			printf("Bad display config block signature (0x%8x)\n", sig); //Azi: issue #48
			return PATCH_ROM_FAILED;
		}
	} else if (dcbtable_version >= 0x14) { /* some NV15/16, and NV11+ */
		char sig[8] = { 0 };
		
		strncpy(sig, (char *)&dcbtable[-7], 7);
		recordlength = 10;

		if (strcmp(sig, "DEV_REC")) {
			printf("Bad Display Configuration Block signature (%s)\n", sig);
			return PATCH_ROM_FAILED;
		}
	} else {
		printf("ERROR: dcbtable_version is 0x%X\n", dcbtable_version);
		return PATCH_ROM_FAILED;
	}
	
	if (numentries >= MAX_NUM_DCB_ENTRIES) {
		numentries = MAX_NUM_DCB_ENTRIES;
	}

	uint8_t num_outputs = 0, i = 0;

	struct dcbentry {
		uint8_t type;
		uint8_t index;
		uint8_t *heads;
	} entries[numentries];

	for (i = 0; i < numentries; i++) {
		uint32_t connection;
		connection = READ_LE_INT(dcbtable,headerlength + recordlength * i);

		/* Should we allow discontinuous DCBs? Certainly DCB I2C tables can be discontinuous */
		if ((connection & 0x0000000f) == 0x0000000f) { /* end of records */ 
			continue;
		}
		if (connection == 0x00000000) { /* seen on an NV11 with DCB v1.5 */ 
			continue;
		}
		if ((connection & 0xf) == 0x6) { /* we skip type 6 as it doesnt appear on macbook nvcaps */
			continue;
		}

		entries[num_outputs].type = connection & 0xf;
		entries[num_outputs].index = num_outputs;
		entries[num_outputs++].heads = (uint8_t*)&(dcbtable[(headerlength + recordlength * i) + 1]);

	}
	
	int has_lvds = false;
	uint8_t channel1 = 0, channel2 = 0;

	for (i = 0; i < num_outputs; i++) {
		if (entries[i].type == 3) {
			has_lvds = true;
			//printf("found LVDS\n");
			channel1 |= ( 0x1 << entries[i].index);
			entries[i].type = TYPE_GROUPED;
		}
	}

	// if we have a LVDS output, we group the rest to the second channel
	if (has_lvds) {
		for (i = 0; i < num_outputs; i++) {
			if (entries[i].type == TYPE_GROUPED) {
				continue;
			}

			channel2 |= ( 0x1 << entries[i].index);
			entries[i].type = TYPE_GROUPED;
		}
	} else {
		int x;
		// we loop twice as we need to generate two channels
		for (x = 0; x <= 1; x++) {
			for (i=0; i<num_outputs; i++) {
				if (entries[i].type == TYPE_GROUPED) {
					continue;
				}
				// if type is TMDS, the prior output is ANALOG
				// we always group ANALOG and TMDS
				// if there is a TV output after TMDS, we group it to that channel as well
				if (i && entries[i].type == 0x2) {
					switch (x) {
						case 0:
							//printf("group channel 1\n");
							channel1 |= ( 0x1 << entries[i].index);
							entries[i].type = TYPE_GROUPED;

							if (entries[i-1].type == 0x0) {
								channel1 |= ( 0x1 << entries[i-1].index);
								entries[i-1].type = TYPE_GROUPED;
							}
							// group TV as well if there is one
							if ( ((i+1) < num_outputs) && (entries[i+1].type == 0x1) ) {
								//	printf("group tv1\n");
								channel1 |= ( 0x1 << entries[i+1].index);
								entries[i+1].type = TYPE_GROUPED;
							}
							break;

						case 1:
							//printf("group channel 2 : %d\n", i);
							channel2 |= ( 0x1 << entries[i].index);
							entries[i].type = TYPE_GROUPED;

							if (entries[i - 1].type == 0x0) {
								channel2 |= ( 0x1 << entries[i-1].index);
								entries[i-1].type = TYPE_GROUPED;
							}
							// group TV as well if there is one
							if ( ((i+1) < num_outputs) && (entries[i+1].type == 0x1) ) {
								//	printf("group tv2\n");
								channel2 |= ( 0x1 << entries[i+1].index);
								entries[i+1].type = TYPE_GROUPED;
							}
							break;
						default:
							break;

					}
					break;
				}
			}
		}
	}
	
	// if we have left ungrouped outputs merge them to the empty channel
	uint8_t *togroup;// = (channel1 ? (channel2 ? NULL : &channel2) : &channel1);
	togroup = &channel2;

	for (i = 0; i < num_outputs; i++) {
		if (entries[i].type != TYPE_GROUPED) {
			//printf("%d not grouped\n", i);
			if (togroup) {
				*togroup |= ( 0x1 << entries[i].index);
			}
			entries[i].type = TYPE_GROUPED;
		}
	}
	
	if (channel1 > channel2) {
		uint8_t buff = channel1;
		channel1 = channel2;
		channel2 = buff;
	}
	
	default_NVCAP[6] = channel1;
	default_NVCAP[8] = channel2;
	
	// patching HEADS
	for (i = 0; i < num_outputs; i++) {
		if (channel1 & (1 << i)) {
			*entries[i].heads = 1;
		} else if(channel2 & (1 << i)) {
			*entries[i].heads = 2;
		}
	}
	return (has_lvds ? PATCH_ROM_SUCCESS_HAS_LVDS : PATCH_ROM_SUCCESS);
}

static char *get_nvidia_model(uint32_t device_id, uint32_t subsys_id)
{
	int i, j;

	// First check in the plist, (for e.g this can override any hardcoded devices)
	cardList_t * nvcard = FindCardWithIds(device_id, subsys_id);
	if (nvcard) {
		if (nvcard->model) {
			return nvcard->model;
		}
	}

	//ErmaC added selector for Chameleon "old" style in System Profiler
//	if (getBoolForKey(kNvidiaGeneric, &showGeneric, &bootInfo->chameleonConfig)) {
//		verbose("\tNvidiaGeneric = Yes\n");

		for (i = 1; i < (sizeof(nvidia_card_generic) / sizeof(nvidia_card_generic[0])); i++) {
			if (nvidia_card_generic[i].device == device_id) {
				return nvidia_card_generic[i].name;
			}
		}
//	return nvidia_card_generic[0].name;
//	}

    
	// Then check the exceptions table
	if (subsys_id) {
		for (i = 0; i < (sizeof(nvidia_card_exceptions) / sizeof(nvidia_card_exceptions[0])); i++) {
			if ((nvidia_card_exceptions[i].device == device_id) && (nvidia_card_exceptions[i].subdev == subsys_id))	{
//				return nvidia_card_exceptions[i].name;
                return nvidia_card_generic[i].name;
				break;
			}
		}
	}

	// At last try the generic names
	for (i = 1; i < (sizeof(nvidia_card_generic) / sizeof(nvidia_card_generic[0])); i++) {
       	if (nvidia_card_generic[i].device == device_id) {
			if (subsys_id) {
				for (j = 0; j < (sizeof(nvidia_card_vendors) / sizeof(nvidia_card_vendors[0])); j++) {
					if (nvidia_card_vendors[j].device == (subsys_id & 0xffff0000)) {
						snprintf(generic_name, 128, "%s %s", // sizeof(generic_name), "%s %s",
							nvidia_card_vendors[j].name, nvidia_card_generic[i].name);
						return &generic_name[0];
					}
				}
			}
			return nvidia_card_generic[i].name;
		}
	}
	return nvidia_card_generic[0].name;
}

static uint32_t load_nvidia_bios_file(const char *filename, uint8_t **buf)
{
	int fd;
	int size;

	if ((fd = open_bvdev("bt(0,0)", filename, 0)) < 0) {
		return 0;
	}

	size = file_size(fd);

	if (size) {
		*buf = malloc(size);
		size = read(fd, (char *)buf, size);
	}
	close(fd);

	return size > 0 ? size : 0;
}

static int devprop_add_nvidia_template(struct DevPropDevice *device)
{
	char tmp[16];

	if (!device)
		return 0;

	if (!DP_ADD_TEMP_VAL(device, nvidia_compatible_0))
		return 0;
	if (!DP_ADD_TEMP_VAL(device, nvidia_device_type_0))
		return 0;
	if (!DP_ADD_TEMP_VAL(device, nvidia_name_0))
		return 0;
	if (!DP_ADD_TEMP_VAL(device, nvidia_compatible_1))
		return 0;
	if (!DP_ADD_TEMP_VAL(device, nvidia_device_type_1))
		return 0;
	if (!DP_ADD_TEMP_VAL(device, nvidia_name_1))
		return 0;
	if (devices_number == 1) {
		if (!DP_ADD_TEMP_VAL(device, nvidia_device_type_parent))
			return 0;
	} else {
		if (!DP_ADD_TEMP_VAL(device, nvidia_device_type_child))
			return 0;
	}

	// Rek : Dont use sprintf return, it does not WORK !! our custom sprintf() always return 0!
	// len = sprintf(tmp, "Slot-%x", devices_number);
	snprintf(tmp, sizeof(tmp), "Slot-%x",devices_number);
	devprop_add_value(device, "AAPL,slot-name", (uint8_t *) tmp, strlen(tmp));
	devices_number++;

	return 1;
}

unsigned long long mem_detect(volatile uint8_t *regs, uint8_t nvCardType, pci_dt_t *nvda_dev, uint32_t device_id, uint32_t subsys_id)
{
	unsigned long long vram_size = 0;

	// First check if any value exist in the plist
	cardList_t * nvcard = FindCardWithIds(device_id, subsys_id);
	if (nvcard) 
	{
		if (nvcard->videoRam > 0) 
		{
			vram_size = nvcard->videoRam * 1024 * 1024;

			return vram_size;
		}
	}

	// Then, Workaround for 9600M GT, GT 210/420/430/440/525M/540M & GTX 560M
	switch (nvda_dev->device_id)
	{
		case 0x0647: // 9600M GT 0647
			vram_size = 512*1024*1024;
			break;
		case 0x0649:	// 9600M GT 0649
			// 10DE06491043202D 1GB VRAM
			if (((nvda_dev->subsys_id.subsys.vendor_id << 16) | nvda_dev->subsys_id.subsys.device_id) == 0x1043202D )
			{
				vram_size = 1024*1024*1024;
			}
			break;
		case 0x0A65: // GT 210
		case 0x0DE0: // GT 440
		case 0x0DE1: // GT 430
		case 0x0DE2: // GT 420
		case 0x0DEC: // GT 525M 0DEC
			vram_size = 1024*1024*1024;
			break;
		case 0x0DE9: // GT 630M
			// 10DE0DE9103C181D 1GB VRAM
			if (((nvda_dev->subsys_id.subsys.vendor_id << 16) | nvda_dev->subsys_id.subsys.device_id) == 0x103C181D )
			{
				vram_size = 1024*1024*1024;
			}
			break;
		case 0x0DF4: // GT 540M
		case 0x0DF5: // GT 525M 0DF5
			vram_size = 1024*1024*1024;
			break;
		case 0x0F00:	// GT 630
			// 10DE0F0014583544 2GB VRAM
			//if (((nvda_dev->subsys_id.subsys.vendor_id << 16) | nvda_dev->subsys_id.subsys.device_id) == 0x14583544 )
			//{
				vram_size = -2147483648UL;//2147483648;
			//}
			break;
		case 0x11C6:	// GTX650TI 11C6
			// 10DE11C61043842A 1GB VRAM
			if (((nvda_dev->subsys_id.subsys.vendor_id << 16) | nvda_dev->subsys_id.subsys.device_id) == 0x1043842A )
			{
				vram_size = 1024*1024*1024;
			}
			break;
		case 0x1251: // GTX 560M
			vram_size = 1536*1024*1024;
			break;
		default:
			break;
	}

	if (!vram_size)	{ // Finally, if vram_size still not set do the calculation with our own method
		if (nvCardType < NV_ARCH_50) {
			vram_size  = REG32(NV04_PFB_FIFO_DATA);
			vram_size &= NV10_PFB_FIFO_DATA_RAM_AMOUNT_MB_MASK;
		} else if (nvCardType < NV_ARCH_C0) {
			vram_size = REG32(NV04_PFB_FIFO_DATA);
			vram_size |= (vram_size & 0xff) << 32;
			vram_size &= 0xffffffff00ll;
		} else { // >= NV_ARCH_C0
			vram_size = REG32(NVC0_MEM_CTRLR_RAM_AMOUNT) << 20;
			vram_size *= REG32(NVC0_MEM_CTRLR_COUNT);
		}
	}

	return vram_size;
}

static bool checkNvRomSig(uint8_t * aRom){
	return aRom != NULL && (aRom[0] == 0x55 && aRom[1] == 0xaa);
}

bool setup_nvidia_devprop(pci_dt_t *nvda_dev)
{
	struct DevPropDevice		*device = NULL;
	char				*devicepath = NULL;
	option_rom_pci_header_t		*rom_pci_header;
	volatile uint8_t		*regs;
	uint8_t				*rom = NULL;
	uint8_t				*nvRom;
	uint8_t				nvCardType = 0;
	unsigned long long		videoRam = 0;
	uint32_t			nvBiosOveride;
	uint32_t			bar[7];
	uint32_t			boot_display = 0;
	int				nvPatch = 0;
	int				len;
	char				biosVersion[64];
	char				nvFilename[64];
	char				kNVCAP[12];
	char				*model = NULL;
	const char			*value;
	bool				doit;

	fill_card_list();

	devicepath = get_pci_dev_path(nvda_dev);
	bar[0] = pci_config_read32(nvda_dev->dev.addr, 0x10 );
	regs = (uint8_t *) (bar[0] & ~0x0f);

	// get card type
	nvCardType = (REG32(0) >> 20) & 0x1ff;

	model = get_nvidia_model(((nvda_dev->vendor_id << 16) | nvda_dev->device_id),((nvda_dev->subsys_id.subsys.vendor_id << 16) | nvda_dev->subsys_id.subsys.device_id));

	// Amount of VRAM in kilobytes
	videoRam = mem_detect(regs, nvCardType, nvda_dev,((nvda_dev->vendor_id << 16) | nvda_dev->device_id),((nvda_dev->subsys_id.subsys.vendor_id << 16) | nvda_dev->subsys_id.subsys.device_id) );

	snprintf(nvFilename, sizeof(nvFilename), "/Extra/%04x_%04x.rom", (uint16_t)nvda_dev->vendor_id, (uint16_t)nvda_dev->device_id);

	if (getBoolForKey(kUseNvidiaROM, &doit, &bootInfo->chameleonConfig) && doit)
	{
		verbose("Looking for nvidia video bios file %s\n", nvFilename);
		nvBiosOveride = load_nvidia_bios_file(nvFilename, &rom);

		if (nvBiosOveride > 0) {
			verbose("Using nVidia Video BIOS File %s (%d Bytes)\n", nvFilename, nvBiosOveride);
			DBG("%s Signature 0x%02x%02x %d bytes\n", nvFilename, rom[0], rom[1], nvBiosOveride);
		} else {
			printf("ERROR: unable to open nVidia Video BIOS File %s\n", nvFilename);
			free(rom);
			return false;
		}
	} else {
		rom = malloc(NVIDIA_ROM_SIZE);
		// Otherwise read bios from card
		nvBiosOveride = 0;

		// PROM first
		// Enable PROM access
		(REG32(NV_PBUS_PCI_NV_20)) = NV_PBUS_PCI_NV_20_ROM_SHADOW_DISABLED;
		nvRom = (uint8_t*)&regs[NV_PROM_OFFSET];

		// Valid Signature ?
		if (checkNvRomSig(nvRom)) {
			bcopy((uint8_t *)nvRom, rom, NVIDIA_ROM_SIZE);
			DBG("PROM Address 0x%x Signature 0x%02x%02x\n", nvRom, rom[0], rom[1]);
		} else {

			// disable PROM access
			(REG32(NV_PBUS_PCI_NV_20)) = NV_PBUS_PCI_NV_20_ROM_SHADOW_ENABLED;

			//PRAM next
			nvRom = (uint8_t*)&regs[NV_PRAMIN_OFFSET];

			if(checkNvRomSig(nvRom)) {
				bcopy((uint32_t *)nvRom, rom, NVIDIA_ROM_SIZE);
				DBG("PRAM Address 0x%x Signature 0x%02x%02x\n", nvRom, rom[0], rom[1]);
			} else {
				// 0xC0000 last
				bcopy((char *)0xc0000, rom, NVIDIA_ROM_SIZE);

				// Valid Signature ?
				if (!checkNvRomSig(rom)) {
					printf("ERROR: Unable to locate nVidia Video BIOS\n");
					return false;
				} else {
                			    DBG("ROM Address 0x%x Signature 0x%02x%02x\n", nvRom, rom[0], rom[1]);
                		}
            		}//end PRAM check
                }//end PROM check
    	}//end load rom from bios

	if ((nvPatch = patch_nvidia_rom(rom)) == PATCH_ROM_FAILED) {
		printf("ERROR: nVidia ROM Patching Failed!\n");
		free(rom);
		//return false;
	}

	rom_pci_header = (option_rom_pci_header_t*)(rom + *(uint16_t *)&rom[24]);

	// check for 'PCIR' sig
	if (rom_pci_header->signature == 0x50434952) {
		if (rom_pci_header->device_id != nvda_dev->device_id) {
			// Get Model from the OpROM
			model = get_nvidia_model(((rom_pci_header->vendor_id << 16) | rom_pci_header->device_id), 0);

			// Get VRAM again
			videoRam = mem_detect(regs, nvCardType, nvda_dev, ((rom_pci_header->vendor_id << 16) | rom_pci_header->device_id), 0);

		} else {
			printf("nVidia incorrect PCI ROM signature: 0x%x\n", rom_pci_header->signature);
		}
	}

    verbose("---------------------------------------------\n");
   	verbose("------------ NVIDIA DEVICE INFO --------------\n");
	verbose("---------------------------------------------\n");
	verbose("%s %dMB NV%02x [%04x:%04x]-[%04x:%04x] :: %s device number: %d\n",
			model, (uint32_t)(videoRam / 1024 / 1024),
			(REG32(0) >> 20) & 0x1ff, nvda_dev->vendor_id, nvda_dev->device_id,
			nvda_dev->subsys_id.subsys.vendor_id, nvda_dev->subsys_id.subsys.device_id,
			devicepath, devices_number);

	if (!string) {
		string = devprop_create_string();
	}
	device = devprop_add_device(string, devicepath);

	/* FIXME: for primary graphics card only */
	boot_display = 1;
	if (devices_number == 1) {
		devprop_add_value(device, "@0,AAPL,boot-display", (uint8_t*)&boot_display, 4);
	}

	if (nvPatch == PATCH_ROM_SUCCESS_HAS_LVDS) {
		uint8_t built_in = 0x01;
		devprop_add_value(device, "@0,built-in", &built_in, 1);
	}

	// get bios version
	const int MAX_BIOS_VERSION_LENGTH = 32;
	char* version_str = (char*)malloc(MAX_BIOS_VERSION_LENGTH);

	memset(version_str, 0, MAX_BIOS_VERSION_LENGTH);

	int i, version_start;
	int crlf_count = 0;

	// only search the first 384 bytes
	for (i = 0; i < 0x180; i++) {
		if (rom[i] == 0x0D && rom[i+1] == 0x0A) {
			crlf_count++;
			// second 0x0D0A was found, extract bios version
			if (crlf_count == 2) {
				if (rom[i-1] == 0x20) i--; // strip last " "

				for (version_start = i; version_start > (i-MAX_BIOS_VERSION_LENGTH); version_start--) {
					// find start
					if (rom[version_start] == 0x00) {
						version_start++;

						// strip "Version "
						if (strncmp((const char*)rom+version_start, "Version ", 8) == 0) {
							version_start += 8;
						}

						strncpy(version_str, (const char*)rom+version_start, i-version_start);
						break;
					}
				}
				break;
			}
		}
	}

	snprintf(biosVersion, sizeof(biosVersion), "%s", (nvBiosOveride > 0) ? nvFilename : version_str);
	snprintf(kNVCAP, sizeof(kNVCAP), "NVCAP_%04x", nvda_dev->device_id);

	if (getValueForKey(kNVCAP, &value, &len, &bootInfo->chameleonConfig) && len == NVCAP_LEN * 2) {
		uint8_t new_NVCAP[NVCAP_LEN];

		if (hex2bin(value, new_NVCAP, NVCAP_LEN) == 0) {
			verbose("Using user supplied NVCAP for %s :: %s\n", model, devicepath);
			memcpy(default_NVCAP, new_NVCAP, NVCAP_LEN);
		}
	}

	if (getValueForKey(kDcfg0, &value, &len, &bootInfo->chameleonConfig) && len == DCFG0_LEN * 2) {
		uint8_t new_dcfg0[DCFG0_LEN];

		if (hex2bin(value, new_dcfg0, DCFG0_LEN) == 0) {
			memcpy(default_dcfg_0, new_dcfg0, DCFG0_LEN);

			verbose("Using user supplied @0,display-cfg\n");
			printf("@0,display-cfg: %02x%02x%02x%02x\n",
				   default_dcfg_0[0], default_dcfg_0[1], default_dcfg_0[2], default_dcfg_0[3]);
		}
	}

	if (getValueForKey(kDcfg1, &value, &len, &bootInfo->chameleonConfig) && len == DCFG1_LEN * 2) {
		uint8_t new_dcfg1[DCFG1_LEN];

		if (hex2bin(value, new_dcfg1, DCFG1_LEN) == 0) {
			memcpy(default_dcfg_1, new_dcfg1, DCFG1_LEN);

			verbose("Using user supplied @1,display-cfg\n");
			printf("@1,display-cfg: %02x%02x%02x%02x\n",
				   default_dcfg_1[0], default_dcfg_1[1], default_dcfg_1[2], default_dcfg_1[3]);
		}
	}

#if DEBUG_NVCAP
	printf("NVCAP: %02x%02x%02x%02x-%02x%02x%02x%02x-%02x%02x%02x%02x-%02x%02x%02x%02x-%02x%02x%02x%02x\n",
	default_NVCAP[0], default_NVCAP[1], default_NVCAP[2], default_NVCAP[3],
	default_NVCAP[4], default_NVCAP[5], default_NVCAP[6], default_NVCAP[7],
	default_NVCAP[8], default_NVCAP[9], default_NVCAP[10], default_NVCAP[11],
	default_NVCAP[12], default_NVCAP[13], default_NVCAP[14], default_NVCAP[15],
	default_NVCAP[16], default_NVCAP[17], default_NVCAP[18], default_NVCAP[19]);
#endif

	devprop_add_nvidia_template(device);
	devprop_add_value(device, "NVCAP", default_NVCAP, NVCAP_LEN);
	devprop_add_value(device, "NVPM", default_NVPM, NVPM_LEN);
	devprop_add_value(device, "VRAM,totalsize", (uint8_t*)&videoRam, 4);
	devprop_add_value(device, "model", (uint8_t*)model, strlen(model) + 1);
	devprop_add_value(device, "rom-revision", (uint8_t*)biosVersion, strlen(biosVersion) + 1);
	devprop_add_value(device, "@0,display-cfg", default_dcfg_0, DCFG0_LEN);
	devprop_add_value(device, "@1,display-cfg", default_dcfg_1, DCFG1_LEN);

	/******************** Added Marchrius.**********************/
	//              For the AppleBacklightDisplay              //
	/***********************************************************/
	if (getBoolForKey(kEnableBacklight, &doit, &bootInfo->chameleonConfig) && doit)	{
		uint8_t AAPL_value[] = {0x01, 0x00, 0x00, 0x00}; //Is the same for all
		devprop_add_value(device, "AAPL,HasPanel", AAPL_value, 4);
		devprop_add_value(device, "AAPL,Haslid", AAPL_value, 4);
		devprop_add_value(device, "AAPL,backlight-control", AAPL_value, 4);
		devprop_add_value(device, "@0,backlight-control", AAPL_value, 4);
	}
	/************************** End ****************************/

	/***********************************************************/
	//                    For the DualLink                     //
	/***********************************************************/
	if (getBoolForKey(kEnableDualLink, &doit, &bootInfo->chameleonConfig) && doit) {
		uint8_t AAPL00_value[] = {0x01, 0x00, 0x00, 0x00};
		devprop_add_value(device, "AAPL00,DualLink", AAPL00_value, 4);
	}
	/************************** End ****************************/

	/************************ HDMI Audio ***********************/
	doit = false;
	//http://forge.voodooprojects.org/p/chameleon/issues/67/
	if(getBoolForKey(kEnableHDMIAudio, &doit, &bootInfo->chameleonConfig) && doit) {
		static uint8_t connector_type_1[]= {0x00, 0x08, 0x00, 0x00};
		devprop_add_value(device, "@0,connector-type",connector_type_1, 4);
		devprop_add_value(device, "@1,connector-type",connector_type_1, 4);
		devprop_add_value(device, "@2,connector-type",connector_type_1, 4);
		devprop_add_value(device, "@3,connector-type",connector_type_1, 4);
	}
	/************************ End Audio *************************/

	if (getBoolForKey(kVBIOS, &doit, &bootInfo->chameleonConfig) && doit) {
		devprop_add_value(device, "vbios", rom, (nvBiosOveride > 0) ? nvBiosOveride : (rom[2] * 512));
	}

	stringdata = malloc(sizeof(uint8_t) * string->length);
	memcpy(stringdata, (uint8_t*)devprop_generate_string(string), string->length);
	stringlength = string->length;

	free(version_str);
	free(rom);
	return true;
}
